/* 
 * \file        AdaptingImageSize.h
 * \author      Juan Cerrón González (juan.cerron@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The definition of AdaptingImageSize method
 *
 * 
 */

#ifndef ADAPTINGIMAGESIZE_H
#define	ADAPTINGIMAGESIZE_H

using namespace std;

class AdaptingImageSize{
public:
    /**
     * Constructor of the class
     * \param h standar proyect's height (480)
     * \param w standar proyect's width (512)
     */
    AdaptingImageSize(int h, int w);
    
    /**
     * Constructor of the class from a vector
     * \param hStd standar proyect's height (480)
     * \param wStd standar proyect's width (512)
     * \param img image in a vector array
     * \param h image's height
     * \param w image's width
     */
    AdaptingImageSize(int hStd, int wStd, unsigned char *img, int h, int w);
    
    ~AdaptingImageSize();
    
    /**
     * It creates a vector from a matrix image
     * \param img matrix
     * \param h matrix height
     * \param w matrix width
     * \return vector
     */
    unsigned char *imageToVector(unsigned char **img, int h, int w);
    
    /**
     * It resizes a fingerprint to the standar dimensions of the proyect.
     * \param image fingerprint to be resized
     * \param h original fingerprint height
     * \param w original fingerprint width
     * \return the resized image
     */
    unsigned char **reSize(unsigned char **image, int h, int w);
    
    /**
     * It completes with white pixels the width or height of the image in order to 
     * maintain equal proportions with the standar
     * \param img original image
     * \param h original image height
     * \param w original image width
     * \return the original image wider
     */
    unsigned char **proportionImage(unsigned char **img, int h, int w);
    
    /**
     * It fills breadthwise the corresponding number of pixels in the image
     * \param img original image
     * \param h original image height
     * \param w original image width
     * \param x proportion between standar widht and height
     * \return the filled image
     */
    unsigned char **fillWidth(unsigned char **img, int h, int w, float x);
          
    /**
     * It fills lengthwise the corresponding number of pixels in the image
     * \param img original image
     * \param h original image height
     * \param w original image width
     * \param x proportion between standar widht and height
     * \return the filled image
     */
    unsigned char **fillHeight(unsigned char **img, int h, int w, float x);
    
    /**
     * It resize a given image to the standar dimensions used in every method
     * \param img image to resize
     * \param h image's height
     * \param w image's width
     * \return the new image resized
     */
    unsigned char **imageSizing(unsigned char **img, int h, int w);
    
    /**
     * It calculates one pixel's neighbours' weight
     * \param indexI pixel row
     * \param indexJ pixel column
     * \param vecinos neighbours
     * \return neighbours weight vector
     */
    void extractWeights(float indexI, float indexJ, int *vecinos, float *pesos);
    
    /**
     * It obtains all pixels that enclose a given pixel
     * \param indexI pixel row
     * \param indexJ pixel column
     * \param h original image's height
     * \param w original image's width
     * \return neighbour pixel vector
     */
    void extractNeighbours(float indexI, float indexJ, int h, int w, int *vecinos);
    
    unsigned char calculateValue(unsigned char **img, int *vecinos, float *pesos);
    
    int gethStd() const;
    int getwStd() const ;
    
    int gethOrig() const;
    int getwOrig() const;
    
    int gethProp() const;
    int getwProp() const;
    
    unsigned char **getImage() const;
public:
    
    int hStd;          //!< Height of the original fingerprint used in the proyect
    int wStd;          //!< Width of the original fingerprint used in the proyect
    
    int hOrig;         //!< Original image height
    int wOrig;         //!< Original image width
    
    int hProp;         //!< Proportioned image height
    int wProp;         //!< Proportioned image width
    
    unsigned char **imagen;    //!< Image used to pass from vector to matrix
    
};

inline int AdaptingImageSize::gethOrig() const { return hOrig;}
inline int AdaptingImageSize::getwOrig() const { return wOrig;}

inline int AdaptingImageSize::gethStd() const { return hStd;}
inline int AdaptingImageSize::getwStd() const { return wStd;}

inline int AdaptingImageSize::gethProp() const { return hProp;}
inline int AdaptingImageSize::getwProp() const { return wProp;}

inline unsigned char **AdaptingImageSize::getImage() const {return imagen;}
#endif	/* ADAPTINGIMAGESIZE_H */

