/* 
 * \file        Cluster.h
 * \author      Juan Cerrón González (juan.cerron@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for Cluster class
 *
 * 
 */

#ifndef CLUSTER_H
#define	CLUSTER_H
    
#include <vector>
#include <math.h>
#include <iostream>

    
using namespace std;    
    
class Cluster {
public:
        int numCores, numDeltas;                // Number of cores and deltas in the cluster
        int blocks;
        vector<float> puntos;                   // Coordinates of all cores an deltas
        float mediaX;                           // Mean coordinate of axis X
        float mediaY;                           // Mean coordinate of axis Y
public:
        /**
         * Cluster's constructor. 
         */ 
        Cluster() {
  
            numCores = 0;
            numDeltas = 0;
            puntos.clear();
            mediaX = 0;
            mediaY = 0;
            blocks = 0;
        }
        
        /**
         * Initializes the object Cluster
         * \param cores  vector with core coordinates
         * \param deltas vector with delta coordinates
         */
        int inicializarCluster(vector<float> cores, vector<float> deltas, int blockCores, int blockDeltas){
            
            numCores = cores.size() / 2;
            numDeltas = deltas.size() / 2;
            
            //Guardamos las coordenadas de los cores
            for (int i = 0; i < numCores; i++){
                puntos.push_back(cores[i * 2]);
                puntos.push_back(cores[i * 2 + 1]);
                blocks += blockCores;
            }
             
            //Guardamos las coordenadas de los deltas
            for (int i = 0; i < numDeltas; i++){
                puntos.push_back(deltas[i * 2]);
                puntos.push_back(deltas[i * 2 + 1]);
                blocks -= blockDeltas;
            }
            
            calcularMedia();
        }

        /**
         * Recorre todas las posiciones pares e impares de "puntos"
         * para calcular la media de la coordenada X e Y.
         */
        int calcularMedia(){
            
            float sumaPar = 0;
            float sumaImpar = 0;
            
            for (int i = 0; i < numCores + numDeltas; i++){
                sumaPar += puntos[i * 2];
                sumaImpar += puntos[i * 2 + 1];
            }
            
            mediaX = sumaPar / (float)(numCores + numDeltas);
            mediaY = sumaImpar / (float)(numCores + numDeltas);
            
            return 0;
        }

        /**
         * Une la información que existe en varios clusteres.
         * \param obj cluster con el que se quiere unir.
         */
        int unirCluster(Cluster obj){
            
            numCores += obj.numCores;
            numDeltas += obj.numDeltas;
            
            blocks += obj.blocks;
            
            //Concatenamos los vectores
            puntos.insert(puntos.end(),obj.puntos.begin(),obj.puntos.end());
            
            calcularMedia();
            
            return 0;
        }
        
        /**
         * 
         * \param obj Cluster to calculate the distance with.
         * \return distance between clusters
         */
        float calcularDistancia(Cluster obj){
            
            float distancia; 
            
            distancia = sqrt(((mediaX) - (obj.mediaX)) * ((mediaX) - (obj.mediaX))
                           + ((mediaY) - (obj.mediaY)) * ((mediaY) - (obj.mediaY)));
            
            return distancia;   
        }
        
        void mostrarElementos(){
            
            for (int i = 0; i < numCores + numDeltas; i++){
                cout << "Elemento "<<i+1<<": " << puntos[i * 2] << ", "  << puntos[i * 2 + 1] << endl;
            }
            
            cout << "Punto Medio: " << mediaX << ", " << mediaY << endl;
        }
};

#endif	/* CLUSTER_H */

