 /* 
 * \file        ComplexFiltersNilssonBigun.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of ComplexFiltersNilssonBigun class
 *
 * 
 */

#include "MemUtil.hpp"
#include "ImgUtil.h"
#include "FingerprintImg.h"

#include "ComplexFiltersNilssonBigun.h"
#include "NewPoincareSingularPoints.h"


#include "ParametersNilssonBigun.h"


ComplexFiltersNilssonBigun::ComplexFiltersNilssonBigun() {
    
    img = 0;
    h = -1;
    w = -1;
    filterResponseX = 0;
    filterResponseY = 0;
    filteredMagnitude = 0;
    filteredAngle = 0;
    coreMaxMag = 0;
    deltaMaxMag = 0;
    segmentation = 0;
    
}

ComplexFiltersNilssonBigun::~ComplexFiltersNilssonBigun() {
    MemUtil::freeMatrix(&img);
    MemUtil::freeMatrix(&segmentation);
    
    if (O != 0) {
        for (int i = 0; i < nLevels; i++) {
            MemUtil::freeMatrix(&O[i]);
        }
        delete[] O;
    }
    if (xCos != 0) {
        for (int i = 0; i < nLevels; i++) {
            MemUtil::freeMatrix(&xCos[i]);
        }
        delete[] xCos;
    }
    if (ySin != 0) {
        for (int i = 0; i < nLevels; i++) {
            MemUtil::freeMatrix(&ySin[i]);
        }
        delete[] ySin;
    }
    if (filterResponseX != 0) {
        for (int t = 0; t < 2; t++) {
            for (int i = 0; i < nLevels; i++) {
                MemUtil::freeMatrix(&filterResponseX[t][i]);
            }
            delete[] filterResponseX[t];
        }
        delete[] filterResponseX;
    }
    if (filterResponseY != 0) {
        for (int t = 0; t < 2; t++) {
            for (int i = 0; i < nLevels; i++) {
                MemUtil::freeMatrix(&filterResponseY[t][i]);
            }
            delete[] filterResponseY[t];
        }
        delete[] filterResponseY;
    }
    if (filteredMagnitude != 0) {
        for (int t = 0; t < 2; t++) {
            for (int i = 0; i < nLevels; i++) {
                MemUtil::freeMatrix(&filteredMagnitude[t][i]);
            }
            delete[] filteredMagnitude[t];
        }
        delete[] filteredMagnitude;
    }
    if (filteredMagnitudeC != 0) {
        for (int t = 0; t < 2; t++) {
            for (int i = 0; i < nLevels; i++) {
                MemUtil::freeMatrix(&filteredMagnitudeC[t][i]);
            }
            delete[] filteredMagnitudeC[t];
        }
        delete[] filteredMagnitudeC;
    }
    if (filteredAngle != 0) {
        for (int t = 0; t < 2; t++) {
            for (int i = 0; i < nLevels; i++) {
                MemUtil::freeMatrix(&filteredAngle[t][i]);
            }
            delete[] filteredAngle[t];
        }
        delete[] filteredAngle;
    }
    cores.clear();
    deltas.clear();
    coreAngles.clear();
    deltaAngles.clear();
}


void ComplexFiltersNilssonBigun::initialize(float **img, int h, int w) {
    this->img = MemUtil::copy2DMatrix(img, h, w);
    this->w = w;
    this->h = h;
}

void ComplexFiltersNilssonBigun::run(string file, PCASYSManager *pcasys, bool segmentationPoincare) {
    
    FingerprintImg fi;

    if (ImgUtil::fAdjust != 1){   

        
		fi.setPCASYSManager(pcasys);        fi.readFingerprintImage(file);
		fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
    
    initialize(fi.getSegmentedImageFloat(), fi.getH(), fi.getW());
    constructGaussianPyramid(sigmaInitialImgNilssonBigun);
    
    if (segmentationPoincare) {
        NewPoincareSingularPoints *npsp = new NewPoincareSingularPoints(img, this->h, this->w);
        npsp->setOrW(orW[nLevels - 1]);
        npsp->setOrH(orH[nLevels - 1]);
        npsp->generateSegmentation(img, this->h, this->w);
        segmentation = MemUtil::copy2DMatrix<bool>(npsp->getSegmentation(), orH[nLevels - 1], orW[nLevels - 1]);
        
        delete npsp;
    }
    
    computeFilterResponses(sigmaFiltersNilssonBigun, sigmag1NilssonBigun, sigmag2PercentageNilssonBigun);
    computeCoresAndDeltas(thresholdCoresNilssonBigun, thresholdDeltasNilssonBigun);
    
}

void ComplexFiltersNilssonBigun::constructGaussianPyramid(float sigma) {
    nLevels = nLevesNilssonBigun;
    int scale = downSampleScaleNilssonBigun;
    /* Filter image with a Gaussian filter, sigma = 0.8 */

    int gW;
    float **gMasc = ImgUtil::createGaussianKernel(sigma, &gW);
    
        
    int hReduced = h - gW + 1;
    int wReduced = w - gW + 1;
    float **filteredImgL0 = ImgUtil::getFilteredImgReduced<float>(img, h, w, gMasc, gW, gW);
    
    if (ImgUtil::DEBUG)
        ImgUtil::writeGrayScaleImage(filteredImgL0, hReduced, wReduced, "gaussianFilteredImg", 1);
    
    float **gradX, **gradY, **grad;
    ImgUtil::computeSobelGradients(filteredImgL0, hReduced, wReduced, &gradX, &gradY, &grad);
    MemUtil::freeMatrix(&filteredImgL0);
    
    if (ImgUtil::DEBUG) {
        string fileName2 = "GradX";
        ImgUtil::writeGrayScaleImage(gradX, hReduced, wReduced, fileName2, 1);
        fileName2 = "GradY";
        ImgUtil::writeGrayScaleImage(gradY, hReduced, wReduced, fileName2, 1);
        fileName2 = "Grad";
        ImgUtil::writeGrayScaleImage(grad, hReduced, wReduced, fileName2, 1);
    }
    
    xCos = new float**[nLevels];
    ySin = new float**[nLevels];    
    O = new float**[nLevels];
    orH = new int[nLevels];
    orW = new int[nLevels];
    int blockW = ceil(blockWNilssonBigun * ImgUtil::fAdjust);
    
    int l = 0;
    
    O[l] = ImgUtil::computeOrientationsFromGradients(gradX, gradY, hReduced, wReduced, blockW, &xCos[l], &ySin[l], &orH[l], &orW[l], false);
        
    MemUtil::freeMatrix(&gradX);
    MemUtil::freeMatrix(&gradY);
    MemUtil::freeMatrix(&grad);
    
    int orHsmooth, orWsmooth;
    float **xCosSmooth = 0, **ySinSmooth = 0;
    for (l = 1; l < nLevels; l++) {
        orHsmooth = orH[l - 1] - gW + 1;
        orWsmooth = orW[l - 1] - gW + 1;
        
        xCosSmooth = ImgUtil::getFilteredImgReduced(xCos[l - 1], orH[l - 1], orW[l - 1], gMasc, gW, gW);
//        MemUtil::freeMatrix(&xCos[l - 1]);
//        xCos[l - 1] = xCosSmooth;
        
        ySinSmooth = ImgUtil::getFilteredImgReduced(ySin[l - 1], orH[l - 1], orW[l - 1], gMasc, gW, gW);
//        MemUtil::freeMatrix(&ySin[l - 1]);
//        ySin[l - 1] = ySinSmooth;
//        
//        orH[l - 1] = orHsmooth;
//        orW[l - 1] = orWsmooth;
        
        xCos[l] = ImgUtil::downSampleImg(xCosSmooth, orHsmooth, orWsmooth, &orH[l], &orW[l], scale);
        ySin[l] = ImgUtil::downSampleImg(ySinSmooth, orHsmooth, orWsmooth, &orH[l], &orW[l], scale);
        
        MemUtil::freeMatrix(&xCosSmooth);
        MemUtil::freeMatrix(&ySinSmooth);
        if (ImgUtil::DEBUG) {
            if (l != 1) {
                string fileName2 = "OrientationSinCosL";
                char str[20];
                sprintf(str, "%d", l);
                fileName2.append(str);            
                ImgUtil::writeOrImg(xCos[l], ySin[l], orW[l], orH[l], fileName2);
            }
        }
    }
    
    MemUtil::freeMatrix(&gMasc);
    
    /* normalize for filtering */
    for (l = 0; l < nLevels; l++) { 
        if (l > 0)
            O[l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        
        for (int i = 0; i < orH[l]; i++) {
            for (int j = 0; j < orW[l]; j++) {                
                O[l][i][j] = 0.5 * atan2(ySin[l][i][j], xCos[l][i][j]);
            }
        }
        
    }
    
}

void ComplexFiltersNilssonBigun::computeFilterResponses(float sigma, float sigmag1, float sigmag2Percentage) {
    
    // hay que construir g(t) y t·g(t) de 1-D
    int gW;
    float *gMasc = ImgUtil::create1DGaussianKernel(sigma, &gW);
    int halfgW = gW / 2;
    
    float sum = 0;
    float *gMascT = MemUtil::createVector<float>(gW);
    for (int k = 0; k < gW; k++) {
        gMascT[k] = gMasc[k] * (k - halfgW);
        sum += abs(gMascT[k]);
    }
    for (int k = 0; k < gW; k++) 
        gMascT[k] /= sum;
    
    filterResponseX = new float***[2];
    filterResponseY = new float***[2];
    filteredMagnitude = new float***[2];
    filteredAngle = new float***[2];
    
    filterResponseX[0] = new float**[nLevels];
    filterResponseY[0] = new float**[nLevels];    
    filteredMagnitude[0] = new float**[nLevels];
    filteredAngle[0] = new float**[nLevels];
    
    filterResponseX[1] = new float**[nLevels];
    filterResponseY[1] = new float**[nLevels];    
    filteredMagnitude[1] = new float**[nLevels];
    filteredAngle[1] = new float**[nLevels];
    
    /* for each level */
    for (int l = 0; l < nLevels; l++) {
        
        filterResponseX[0][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filterResponseY[0][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filteredMagnitude[0][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filteredAngle[0][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filterResponseX[1][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filterResponseY[1][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filteredMagnitude[1][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        filteredAngle[1][l] = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        
        
        float **fRX_xCos = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        float **fRX_ySin = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        float **fRY_xCos = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        float **fRY_ySin = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        /* for each position */
        for (int i = halfgW; i < orH[l] - halfgW  ; i++) {
            for (int j = halfgW; j < orW[l] - halfgW ; j++) {
                
                /* pasamos el fltro t·g(t) */
                for (int k = 0; k < gW; k++) {
                    int v = j + (k - halfgW);                 
                    fRX_xCos[i][j] += gMascT[k] * cos(2. * O[l][i][v]); 
                    fRX_ySin[i][j] += gMascT[k]  * sin(2. * O[l][i][v]);
                                     
                    int u = i + (k - halfgW);      
                    fRY_xCos[i][j] += gMascT[k]  * cos(2. * O[l][u][j]); 
                    fRY_ySin[i][j] += gMascT[k]  * sin(2. * O[l][u][j]); 
                }
            
            }
        }
        float **fRX_xCos_2 = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        float **fRX_ySin_2 = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        float **fRY_xCos_2 = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        float **fRY_ySin_2 = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
        
        for (int i = halfgW; i < orH[l] - halfgW; i++) {
            for (int j = halfgW; j < orW[l] - halfgW ; j++) {
                
                /* pasamos el fltro g(t) */
                for (int k = 0; k < gW; k++) {
                    int u = i + (k - halfgW);     
                    fRX_xCos_2[i][j] += gMasc[k] * fRX_xCos[u][j];
                    fRX_ySin_2[i][j] += gMasc[k] * fRX_ySin[u][j];
                                                      
                    int v = j + (k - halfgW);
                    fRY_xCos_2[i][j] += gMasc[k] * fRY_xCos[i][v];
                    fRY_ySin_2[i][j] += gMasc[k] * fRY_ySin[i][v];
                }
                filterResponseX[0][l][i][j] = fRX_xCos_2[i][j] - fRY_ySin_2[i][j];
                filterResponseY[0][l][i][j] = fRY_xCos_2[i][j] + fRX_ySin_2[i][j];
                // magnitude y luego ángulo
                //2. / (float)(gW * gW) 1.  / (float)(gW) * 
                filteredMagnitude[0][l][i][j] = sqrt(filterResponseX[0][l][i][j] * filterResponseX[0][l][i][j] + 
                        filterResponseY[0][l][i][j] * filterResponseY[0][l][i][j]) / 2.; //sqrt(4);
                filteredAngle[0][l][i][j] = atan2(filterResponseY[0][l][i][j], filterResponseX[0][l][i][j]);
                
                filterResponseX[1][l][i][j] = fRX_xCos_2[i][j] + fRY_ySin_2[i][j];
                filterResponseY[1][l][i][j] = -fRY_xCos_2[i][j] + fRX_ySin_2[i][j];
                // magnitude y luego ángulo 2. / (float)(gW * gW) *  
                filteredMagnitude[1][l][i][j] = sqrt(filterResponseX[1][l][i][j] * filterResponseX[1][l][i][j] + 
                        filterResponseY[1][l][i][j] * filterResponseY[1][l][i][j]) / 2.; //sqrt(4);
                filteredAngle[1][l][i][j] = atan2(filterResponseY[1][l][i][j], filterResponseX[1][l][i][j]);
            
            }
        }
        
        MemUtil::freeMatrix(&fRX_xCos);
        MemUtil::freeMatrix(&fRX_ySin);
        MemUtil::freeMatrix(&fRY_xCos);
        MemUtil::freeMatrix(&fRY_ySin);
        
        // MEJOr MEMSET!!! Y PONER A 0
        MemUtil::freeMatrix(&fRX_xCos_2);
        MemUtil::freeMatrix(&fRX_ySin_2);
        MemUtil::freeMatrix(&fRY_xCos_2);
        MemUtil::freeMatrix(&fRY_ySin_2);
        
        
        if ((ImgUtil::DEBUG)) {
            string fileName2 = "FilterResponseMagnitudeCore_L";
            char str[20];
            sprintf(str, "%d", l);
            fileName2.append(str);           
            ImgUtil::writeGrayScaleImage(filteredMagnitude[0][l], orH[l], orW[l], fileName2, 255);
            fileName2 = "FilterResponseMagnitudeDelta_L";
            sprintf(str, "%d", l);
            fileName2.append(str);           
            ImgUtil::writeGrayScaleImage(filteredMagnitude[1][l], orH[l], orW[l], fileName2, 255);
        }
    }
    
    MemUtil::freeMatrix(&gMasc);
    MemUtil::freeMatrix(&gMascT);
    
    /* Level nLevels - 1 processing...*/
    gMasc = ImgUtil::create1DGaussianKernel(sigmag1, &gW);
    halfgW = gW / 2;
    
    
    int l = nLevels - 1;
    float **g1convz3_1 = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
    float **g1convz3 = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
    float **magni = MemUtil::create2DMatrix<float>(orH[l], orW[l]);
    float magmax = -1;
    for (int i = halfgW; i < orH[l] - halfgW; i++) {
        for (int j = halfgW; j < orW[l] - halfgW; j++) {            
            for (int k = 0; k < gW; k++) {
                float mag = sqrt(xCos[l][i + (k - halfgW)][j] * xCos[l][i + (k - halfgW)][j] + ySin[l][i + (k - halfgW)][j] * ySin[l][i + (k - halfgW)][j]);
                g1convz3_1[i][j] += mag * gMasc[k];   
            }
            magni[i][j] = sqrt(xCos[l][i][j] * xCos[l][i][j] + ySin[l][i][j] * ySin[l][i][j]);
            if (magni[i][j] > magmax)
                magmax = magni[i][j];
        }
    }
    for (int i = halfgW; i < orH[l] - halfgW; i++) {
        for (int j = halfgW; j < orW[l] - halfgW; j++) {            
            for (int k = 0; k < gW; k++)
                g1convz3[i][j] += g1convz3_1[i][j + (k - halfgW)] * gMasc[k];      
            g1convz3[i][j] /= magmax;
            magni[i][j] /= magmax;
        }
    }
    if ((ImgUtil::DEBUG)) {
            string fileName2 = "g1convz3";
            char str[20];
            sprintf(str, "%d", l);
            fileName2.append(str);           
            ImgUtil::writeGrayScaleImage(g1convz3, orH[l], orW[l], fileName2, 255);
            
            
            fileName2 = "magni";
            ImgUtil::writeGrayScaleImage(magni, orH[l], orW[l], fileName2, 255);
    }
    MemUtil::freeMatrix(&magni);
    
    float **gMasc2 = ImgUtil::createGaussianKernel(sigmag2Percentage * orH[l], sigmag2Percentage * orW[l] , orH[l], orW[l]);
    
    float maxMasc = gMasc2[orH[l] / 2][orW[l] / 2];
    for (int i = 0; i < orH[l]; i++) {
        for (int j = 0; j < orW[l]; j++) {            
            gMasc2[i][j] /= maxMasc;
            filteredMagnitude[0][l][i][j] = 0.5 * (filteredMagnitude[0][l][i][j] * g1convz3[i][j] + filteredMagnitude[0][l][i][j] * gMasc2[i][j]);
            filteredMagnitude[1][l][i][j] = 0.5 * (filteredMagnitude[1][l][i][j] * g1convz3[i][j] + filteredMagnitude[1][l][i][j] * gMasc2[i][j]);
        }
    }
    
    if ((ImgUtil::DEBUG)) {
        string fileName2 = "FilterResponseMagnitudeCore_L3processed";
        ImgUtil::writeGrayScaleImage(filteredMagnitude[0][l], orH[l], orW[l], fileName2, 255);
        fileName2 = "FilterResponseMagnitudeDelta_L3processed";
        ImgUtil::writeGrayScaleImage(filteredMagnitude[1][l], orH[l], orW[l], fileName2, 255);
        fileName2 = "gMasc2";
        ImgUtil::writeGrayScaleImage(gMasc2, orH[l], orW[l], fileName2, 255);
    }
    
    MemUtil::freeMatrix(&g1convz3_1);
    MemUtil::freeMatrix(&g1convz3);
    MemUtil::freeMatrix(&gMasc);
    MemUtil::freeMatrix(&gMasc2);
    
    
    filteredMagnitudeC = new float***[2];
    filteredMagnitudeC[0] = new float**[nLevels];
    filteredMagnitudeC[1] = new float**[nLevels];
    
    /* for each level */
    for (int l = 0; l < nLevels; l++) {        
        filteredMagnitudeC[0][l] = MemUtil::copy2DMatrix<float>(filteredMagnitude[0][l], orH[l], orW[l]);
        filteredMagnitudeC[1][l] = MemUtil::copy2DMatrix<float>(filteredMagnitude[1][l], orH[l], orW[l]);
    }
        
    float aux;
    /* convertir a s ! */
    for (int l = 0; l < nLevels; l++) {        
        /* for each position */
        for (int i = halfgW; i < orH[l] - halfgW; i++) {
            for (int j = halfgW; j < orW[l] - halfgW; j++) {
                aux = filteredMagnitude[0][l][i][j];
                filteredMagnitude[0][l][i][j] = filteredMagnitude[0][l][i][j] * (1. - filteredMagnitude[1][l][i][j]);
                filteredMagnitude[1][l][i][j] = filteredMagnitude[1][l][i][j] * (1. - aux);
            }
        }
        if ((ImgUtil::DEBUG)) {
            string fileName2 = "FilterResponseMagnitudeCoreS_L";
            char str[20];
            sprintf(str, "%d", l);
            fileName2.append(str);           
            ImgUtil::writeGrayScaleImage(filteredMagnitude[0][l], orH[l], orW[l], fileName2, 255);
            fileName2 = "FilterResponseMagnitudeDeltaS_L";
            sprintf(str, "%d", l);
            fileName2.append(str);           
            ImgUtil::writeGrayScaleImage(filteredMagnitude[1][l], orH[l], orW[l], fileName2, 255);
        }
    }
                
}


void ComplexFiltersNilssonBigun::computeCoresAndDeltas(float thresholdCores, float thresholdDeltas) {
    
    /* Find local maximuns in nLevels - 1 */
    int l = nLevels - 1;
    vector<int> localMaximumCores = findLocalMaximum(filteredMagnitude[0][l], orH[l], orW[l], thresholdCores);
    vector<int> localMaximumDeltas = findLocalMaximum(filteredMagnitude[1][l], orH[l], orW[l], thresholdDeltas);
    
    /* in a 13x13 window, find whther it passes the threshold! sino guardar 1 al menos en cores!! */
    
    float *magnitud = MemUtil::createVector<float>(1);
    
    if (ImgUtil::DEBUG)
        cout << "goto cores" << endl;
    vector<int> coresLevel2 = findInLowerLevel(localMaximumCores, filteredMagnitude[0][l - 1], orH[l - 1], orW[l - 1], thresholdCores, wSizeMaxSearchNilssonBiggun, magnitud, true);
    coreMaxMag = magnitud[0];
    
    if (ImgUtil::DEBUG) {
        cout << "MAGNITUD = " << coreMaxMag << endl;
        cout << "goto deltas" << endl;
    }
    vector<int> deltasLevel2 = findInLowerLevel(localMaximumDeltas, filteredMagnitude[1][l - 1], orH[l - 1], orW[l - 1], thresholdDeltas, wSizeMaxSearchNilssonBiggun, magnitud);
    deltaMaxMag = magnitud[0];
    if (ImgUtil::DEBUG)
        cout << "MAGNITUD = " << deltaMaxMag << endl;
    
    
    
    /* Calcular el ángulo! */
    if (ImgUtil::DEBUG)
        cout << "goto cores angle" << endl;
    coreAngles = computeAngles(coresLevel2, 0);
    if (ImgUtil::DEBUG)
        cout << "goto deltas angle" << endl;
    deltaAngles = computeAngles(deltasLevel2, 1);
    
    /* Buscar la posición en un nivel más abajo */
    if (ImgUtil::DEBUG)
        cout << "goto cores" << endl;
    vector<int> coresLevel1 = findInLowerLevel(coresLevel2, filteredMagnitude[0][l - 2], orH[l - 2], orW[l - 2], 0, wSizeMaxSearchNilssonBiggun, magnitud, true);
    if (ImgUtil::DEBUG)
        cout << "goto deltas" << endl;
    vector<int> deltasLevel1 = findInLowerLevel(deltasLevel2, filteredMagnitude[1][l - 2], orH[l - 2], orW[l - 2], 0, wSizeMaxSearchNilssonBiggun, magnitud);
    
    
    /* Posición exacta en pixeles en nivel 1! */
    cores = computePixelPosition(coresLevel1);
    deltas = computePixelPosition(deltasLevel1);
    
    if (ImgUtil::DEBUG) {
        string fileName = "coresAndDeltasNilssonBigun";
        ImgUtil::writeSPImage(img, h, w, deltas, cores, 1, fileName);
    }

    MemUtil::freeMatrix(&magnitud);
}

vector<int> ComplexFiltersNilssonBigun::findLocalMaximum(float **matrix, int h, int w, float threshold) {
    
    vector<int> localMaximum;
    int iAbsMax = -1;
    int jAbsMax = -1;
    float absMax = -1;
    
    for (int i = 1; i < h - 1; i++) {
        for (int j = 1; j < w - 1; j++) {
            bool notMax = false;
//            cout << matrix[i][j] << endl;
            if ((segmentation == 0) || (segmentation[i][j] == true) ) {
                if ((matrix[i][j] >= threshold / 4)) { // && ((segmentation == 0) || (segmentation[i][j] == true) )) { //2.5) { // es para quitar algunos de más, el threshold se usa en el siguiente nivel
                    for (int u = i - 1; ((u <= i + 1) && (!notMax)); u++) {
                        for (int v = j - 1; v <= j + 1; v++) {
                            if ((matrix[u][v] > matrix[i][j])) { //  && ((segmentation == 0) || (segmentation[u][v]))) {
                                notMax = true;
                                break;
                            }
    //                        cout << matrix[u][v] << " ";
                        }
    //                    cout << endl;
                    }
    //                    cout << endl;
    //                    cout << endl;
                    if (!notMax) {
                        if (ImgUtil::DEBUG)
                            cout << "localMaximum in " << i << ", " << j << ": " << matrix[i][j] << endl;
                        localMaximum.push_back(i);
                        localMaximum.push_back(j);
                    }
                }
                if ((matrix[i][j] > absMax)) { // && ((segmentation == 0) || (segmentation[i][j]) )) {
                    absMax = matrix[i][j];
                    iAbsMax = i;
                    jAbsMax = j;
                }
            }
            
        }
    }
    if (localMaximum.size() == 0) {
        if (ImgUtil::DEBUG)
            cout << "Absolute Max needed! in " << iAbsMax << ", " << jAbsMax << endl;
        localMaximum.push_back(iAbsMax);
        localMaximum.push_back(jAbsMax);
    }
    if (ImgUtil::DEBUG)
        cout << "Absolute Max = " << absMax << endl;
    return localMaximum;
    
}


vector<int> ComplexFiltersNilssonBigun::findInLowerLevel(vector<int> localMaximum, float **matrix, int h, int w, float threshold, int s, float *magnitud, bool core) {
    
    vector<int> singularPoints;
    
    magnitud[0] = -100000;
    int iAbsMax = -1;
    int jAbsMax = -1;
    float absMax = -1;
    
    float localMax = -1;
    int iLocalMax = -1;
    int jLocalMax = -1;
    /* For each localMaximum, find the maximum in 2s + 1 x 2s + 1 window, if it does not exist and passes the threshold, it si considered a core/delta */
    // if localMaximum ends, and no SP was found, the absolute maximum must be used!
    while (!localMaximum.empty()) {
        int j = localMaximum.back() * downSampleScaleNilssonBigun + ceil(sigmaFiltersNilssonBigun * 3.);    // transform coordinates!!!!
        localMaximum.pop_back();
        int i = localMaximum.back() * downSampleScaleNilssonBigun + ceil(sigmaFiltersNilssonBigun * 3.);
        localMaximum.pop_back();
        
        localMax = matrix[i][j];
        iLocalMax = i;
        jLocalMax = j;
        for (int u = max(0, i - s); u <= min(i + s, h - 1); u++) {
            for (int v = max(0, j - s); v <= min(j + s, w - 1); v++) {
                if (matrix[u][v] > localMax) {
                    localMax = matrix[u][v];
                    iLocalMax = u;
                    jLocalMax = v;
                }
            }
        }
        int iAux = (int)((iLocalMax -  ceil(sigmaFiltersNilssonBigun * 3.)) / downSampleScaleNilssonBigun);
        int jAux = (int)((jLocalMax -  ceil(sigmaFiltersNilssonBigun * 3.)) / downSampleScaleNilssonBigun);
        if ((threshold == 0) || (segmentation == 0) || ((iAux >= 0) && (jAux >= 0) && (segmentation[iAux][jAux] ))) {
            if (ImgUtil::DEBUG)
                    cout << "Local Max = " << localMax << endl;
            if (localMax >= threshold) {
                bool notIn = true;
                for (int k = 0; k < singularPoints.size() / 2; k++) {
                    if ((singularPoints[2 * k] == iLocalMax) && (singularPoints[2 * k + 1] == jLocalMax))
                        notIn = false;                        
                }
                if (notIn) {
                    singularPoints.push_back(iLocalMax);
                    singularPoints.push_back(jLocalMax);
                    magnitud[0] = (localMax > magnitud[0]) ? localMax : magnitud[0];
                    if (ImgUtil::DEBUG)
                        cout << "Singular Point in " << iLocalMax << ", " << jLocalMax << ": " << localMax << endl;
                }
            }
            if (localMax > absMax) {
                absMax = localMax;
                iAbsMax = iLocalMax;
                jAbsMax = jLocalMax;
            }
        }
    }
    if (singularPoints.empty() && core) {
        singularPoints.push_back(iAbsMax);
        singularPoints.push_back(jAbsMax);
        magnitud[0] = absMax;
        if (ImgUtil::DEBUG)
            cout << "Absolute Max Needed!! -> Singular Point in " << iAbsMax << ", " << jAbsMax << ": " << absMax << endl;
    }
    
    return singularPoints;
    
}

vector<int> ComplexFiltersNilssonBigun::computePixelPosition(vector<int> singularPoints) {
    
    vector<int> pixelPosition;
    while(!singularPoints.empty()) {
        int j = singularPoints.back() * downSampleScaleNilssonBigun + ceil(sigmaFiltersNilssonBigun * 3.);    // transform coordinates!!!!
        singularPoints.pop_back();
        int i = singularPoints.back() * downSampleScaleNilssonBigun + ceil(sigmaFiltersNilssonBigun * 3.);
        singularPoints.pop_back();
        
        if ((i >= 4) && (i < h - 4) && (j >= 4) && (j < w - 4)) {
            pixelPosition.push_back(i);
            pixelPosition.push_back(j);
            if (ImgUtil::DEBUG)
                cout << "Singular Point in pixels = " << i << ", " << j << endl;
        }
    }
    
    return pixelPosition;
}

vector<float> ComplexFiltersNilssonBigun::computeAngles(vector<int> singularPoints, int type) {
    
    vector<float> angles;
    
    float w[3][3];
    float wUp[3][3];
    float dSS = downSampleScaleNilssonBigun;
    int width =  ceil(sigmaFiltersNilssonBigun * 3.);
    
    for (int k = 0; k < singularPoints.size(); k += 2) {
        int i = singularPoints[k];
        int j = singularPoints[k + 1];
        
        
        float magSum = 0;
        float magSumUp = 0;
        for (int u = max(width, i - 1); u <= min(i + 1, orH[nLevels - 2] - width); u++) {
            for (int v = max(width, j - 1); v <= min(j + 1, orW[nLevels - 2] - width); v++) {
                magSum += filteredMagnitudeC[type][nLevels - 2][u][v]; 
                magSumUp += filteredMagnitudeC[type][nLevels - 1][(int)((u - width) / dSS)][(int)((v - width) / dSS)]; 
                
            }
        }
        
        if (ImgUtil::DEBUG)
            cout << "magSum = " << magSum << " magSumUp = " << magSumUp << endl;
        
        float sumX = 0, sumY = 0;
        float sumXUp = 0, sumYUp = 0;
        for (int u = max(width, i - 1); u <= min(i + 1, orH[nLevels - 2] - width); u++) {
            for (int v = max(width, j - 1); v <= min(j + 1, orW[nLevels - 2] - width); v++) {
                int i2 = u - i + 1;
                int j2 = v - j + 1;
                int uUp = (u - width) / dSS;
                int vUp = (v - width) / dSS;
                if (magSum != 0)
                    w[i2][j2] = filteredMagnitudeC[type][nLevels - 2][u][v] / magSum; 
                else
                    w[i2][j2] = 0;
                if (magSumUp != 0)
                    wUp[i2][j2] = filteredMagnitudeC[type][nLevels - 1][uUp][vUp] / magSumUp; 
                else
                    wUp[i2][j2] = 0;
                sumX += w[i2][j2] * filterResponseX[type][nLevels - 2][u][v];
                sumY += w[i2][j2] * filterResponseY[type][nLevels - 2][u][v];
                
                
                sumXUp += wUp[i2][j2] * filterResponseX[type][nLevels - 1][uUp][vUp];
                sumYUp += wUp[i2][j2] * filterResponseY[type][nLevels - 1][uUp][vUp];
            }
        }
        sumX += sumXUp;
        sumY += sumYUp;
        if (ImgUtil::DEBUG)
            cout << "sumX = " << sumX << " sumY = " << sumY << endl;
        float ang = atan2(sumY, sumX);
        if (type == 1)
            ang /= 3.;
        if (ImgUtil::DEBUG)
            cout << "angulo = " << ang * 180. / PI << endl;
        angles.push_back(ang);
    }
    
    
    return angles;
}