/* 
 * \file        EejcutarMetodo.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Juan Cerrón González (juan.cerron@unavarra.es)
 * \version     1.0
 * 
 * \SECTION     
 * 
 * 
 * It creates the Data Set file using the fingerprints of a directory.
 * The file includes a header with the attributes' names and maximum
 * and minimum values.
 * 
 * 
 */
#include <sstream>
#include <algorithm>

#include "EjecutarMetodo.h"

#include "LlamaMetodos.h"


EjecutarMetodo::EjecutarMetodo(){

    maximos = 0;
    minimos = 0;
    
    tipos = 0;
    valoresNom = 0;

    fTemporal = "";
}

EjecutarMetodo::~EjecutarMetodo(){

    MemUtil::freeMatrix(&maximos);
    MemUtil::freeMatrix(&minimos);
    
    MemUtil::freeMatrix(&tipos);
    MemUtil::freeMatrix(&valoresNom);

}

int EjecutarMetodo::getdir (string dir, vector<string> &files)
{
    DIR *dp;
    struct dirent *dirp;
    if((dp = opendir(dir.c_str())) == NULL) {
        // cout << "Error opening " << dir << endl;
        return -1;
    }

    while ((dirp = readdir(dp)) != NULL) {
        if ((string(dirp->d_name) != ".") && (string(dirp->d_name) != "..") && string(dirp->d_name).find(".wsq_") == string(dirp->d_name).npos && string(dirp->d_name).find(".wsq") != string(dirp->d_name).npos)
                files.push_back(string(dirp->d_name));
    }
    closedir(dp);
    sort(files.begin(), files.end());
    return 0;
    
}

int EjecutarMetodo::getDB(string dir){
    
    size_t found,found2;
    
    found=dir.find("DB");
    found2=dir.find("db");
    if (found!=string::npos){
        ImgUtil::setCurrentDB(dir.substr(found,found+3));
        return 0;
    }
    else if (found2!=string::npos){
        ImgUtil::setCurrentDB(dir.substr(found2,found2+3));
        return 0;
    }
    else{
        ImgUtil::setCurrentDB("DBX");
        return -1;
    }
    
    
}

void EjecutarMetodo::printFeatureVector(float *fv, int n, string file, char *metodo, ostream &out) {
   
    string name = metodo;
    size_t foundKaru, foundZhang, foundWang;
    
    out.setf ( std::ios_base::uppercase );
    
    size_t foundLeung;
    foundLeung = name.find("Leung");
    bool esLeung = (foundLeung != string::npos)? true : false;
    
    //Para escribir el tipo de huella como letra en lugar de como número
    foundKaru = name.find("Karu");
    foundZhang = name.find("Zhang");
    foundWang = name.find("Wang");
    if (foundKaru != string::npos || foundZhang != string::npos || foundWang != string::npos){

        if (n != 1){
            if (fv[0] == 1)
                out << "A, ";
            else if (fv[0] == 2)
                out << "T, ";
            else if (fv[0] == 3)
                out << "R, ";
            else if (fv[0] == 4)
                out << "L, ";
            else
                out << "W, ";

            for (int i = 1; i < n - 1; i++){
                out << fv[i] << ", ";
            }
            out << fv[n - 1];
        }
        else{
            if (fv[0] == 1)
                out << "A";
            else if (fv[0] == 2)
                out << "T";
            else if (fv[0] == 3)
                out << "R";
            else if (fv[0] == 4)
                out << "L";
            else if (fv[0] == 0)
                out << "W";
            else
		out << "?";
        }
    }
    else{
        if (n > 1){
            if (fv[0] == -1 && esLeung)
                out << "?, ";
            else
                out << fv[0] << ", ";
            
            for (int i = 1; i < n - 1; i++){
                if (fv[i] == -1 && esLeung)
                    out << "?, ";
                else
                    out << fv[i] << ", ";
            }

            if (fv[n - 1] == -1 && esLeung)
                out << "?";
            else
                out << fv[n - 1];
        }
        else
            if (fv[0] == -1 && esLeung)
                out << "?";
            else
                out << fv[0];

    } 
    
    int type = extractFingerprintType(file);
    if (type != -1){
        
        if (type == 1)
            out << ", A";
        else if (type == 2)
            out << ", T";
        else if (type == 3)
            out << ", R";
        else if (type == 4)
            out << ", L";
        else
            out << ", W";
        
        if (searchElto(valoresNom[n], type) == -1){
            valoresNom[n][++valoresNom[n][0]] = type;
        }
    }
    
    out << endl;
    
    for (int i = 0; i < n; i++){
        //Si es un numero real guardamos el máximo y mínimo de cada atributo
        if ((fv[i] != -1 || !esLeung) && (((int) fv[i]) != fv[i] || tipos[i] == 1)){
            if (maximos[i] < fv[i])
                maximos[i] = fv[i];
        
            if (minimos[i] > fv[i])
                minimos[i] = fv[i];
            
            tipos[i] = 1; //Indicamos que es real
        }
        //Si es un número entero, entendemos que es nominal y guardamos
        //el valor que corresponde además de añadirlo (si no está) a la 
        //lista de posibles valores nominales
        else if (fv[i] != -1 || !esLeung){
            //Por si al principio coincide que a pesar de ser real, el número obtenido 
            //es de la forma 1,00000. Debemos guardar este número en máximos y mínimos.
            if (maximos[i] < fv[i])
                maximos[i] = fv[i];
        
            if (minimos[i] > fv[i])
                minimos[i] = fv[i];
   
            //Si no está el valor nominal del atributo en la lista de nominales, se guarda 
            //y se vuelve a insertar en la lista en el índice correspondiente.
            if (searchElto(valoresNom[i], (int) fv[i]) == -1){
                if (valoresNom[i][0] + 1 <= MaxValNom) {
                    valoresNom[i][++valoresNom[i][0]] = fv[i];
                }
                else
                    tipos[i] = 1;
            }
        }
    }
}

void EjecutarMetodo::printDataSetHeader(int n, char *metodo, ostream &out){
    
    out.setf ( std::ios_base::uppercase );
    //Nombre del data set
    out << "@relation FingerprintDataSet" << endl;
    
    //Atributos
    string name = metodo;
    size_t foundKaru, foundZhang, foundWang;
   
    foundKaru = name.find("Karu");
    foundZhang = name.find("Zhang");
    foundWang = name.find("Wang");
    if (foundKaru != string::npos || foundZhang != string::npos || foundWang != string::npos){
        
        out << "@attribute A1";
        
        orderVector(valoresNom[0]);

        out << " {";

        for (int j = 1; j < valoresNom[0][0]; j++){
            if (valoresNom[0][j] == 1)
                out << "A, ";
            else if (valoresNom[0][j] == 2)
                out << "T, ";
            else if (valoresNom[0][j] == 3)
                out << "R, ";
            else if (valoresNom[0][j] == 4)
                out << "L, ";
            else
                out << "W, ";
        }
        if (valoresNom[0][valoresNom[0][0]] == 1)
            out << "A}" << endl;
        else if (valoresNom[0][valoresNom[0][0]] == 2)
            out << "T}" << endl;
        else if (valoresNom[0][valoresNom[0][0]] == 3)
            out << "R}" << endl;
        else if (valoresNom[0][valoresNom[0][0]] == 4)
            out << "L}" << endl;
        else
            out << "W}" << endl;    
        
    }
    else{
        out << "@attribute A1";
        
        if (tipos[0] == 1){
            out << " real [" << minimos[0] << "," << maximos[0] << "]" << endl;
        }
        else if (tipos[0] == 0){
            
            orderVector(valoresNom[0]);
            
            out << " {";
            
            for (int j = 1; j < valoresNom[0][0]; j++){
                out << valoresNom[0][j] << ", ";
            }
            
            out << valoresNom[0][valoresNom[0][0]] << "}" << endl;
        }
    }
    
    for (int i = 1; i < n; i++){
        
        out << "@attribute A" << i + 1;
        
        if (tipos[i] == 1){
            out << " real [" << minimos[i] << "," << maximos[i] << "]" << endl;
        }
        else if (tipos[i] == 0){
            
            orderVector(valoresNom[i]);
            
            out << " {";
            
            for (int j = 1; j < valoresNom[i][0]; j++){
                out << valoresNom[i][j] << ", ";
            }
            
            out << valoresNom[i][valoresNom[i][0]] << "}" << endl;
        }
        
    }
    
    if (valoresNom[n][0] != 0){
        
        out << "@attribute Class {";
        for (int i = 1; i < valoresNom[n][0]; i++){
            if (valoresNom[n][i] == 1)
                out << "A, ";
            else if (valoresNom[n][i] == 2)
                out << "T, ";
            else if (valoresNom[n][i] == 3)
                out << "R, ";
            else if (valoresNom[n][i] == 4)
                out << "L, ";
            else if (valoresNom[n][i] == 5)
                out << "W, ";
        }
        
        if (valoresNom[n][valoresNom[n][0]] == 1)
            out << "A}" << endl;
        else if (valoresNom[n][valoresNom[n][0]] == 2)
            out << "T}" << endl;
        else if (valoresNom[n][valoresNom[n][0]] == 3)
            out << "R}" << endl;
        else if (valoresNom[n][valoresNom[n][0]] == 4)
            out << "L}" << endl;
        else if (valoresNom[n][valoresNom[n][0]] == 5)
            out << "W}" << endl;
    }
    
    //Nombre de entradas
    out << "@inputs ";
    for (int i = 0; i < n - 1; i++){
        out << "A" << i + 1 << ", ";
    }
    out << "A" << n << endl;
    
    if (valoresNom[n][0] != 0)
        out << "@outputs Class" << endl;
    
    
    //Etiqueta que indica el inicio de los datos
    out << "@data" << endl;
}
void EjecutarMetodo::orderVector(int *v){
    
    int aux = 0;
    
    for (int i = 2; i <= v[0]; i++){
        for (int j = 1; j <=  v[0] - 1; j++){
            if (v[j] > v [j + 1]){
                aux = v[j];
                v[j] = v[j + 1];
                v[j + 1] = aux;
            }
        }
    }
}

int EjecutarMetodo::searchElto(int *v, int valor){
    
    for (int i = 1; i <= v[0]; i++) //Porque suponemos q como mucho habrá 10 valores nominales distintos en nuestros algoritmos
        if (v[i] == valor)
            return 0;
    return -1;
}

void EjecutarMetodo::copyDataSet(ostream &out){
    ifstream fTemp;
    string linea;
    fTemp.open(fTemporal.c_str(), ifstream::in);
    
    while (getline(fTemp, linea))
        out << linea << endl;
    
    fTemp.close();
}

void EjecutarMetodo::ejecutar(char *directorio, char *metodo, char *fichSalida, bool ajustar, bool realzar, bool fichSalidaOk, bool dbg){
    
    ImgUtil::setDebug(dbg);
    
    vector<string> files = vector<string>();
    float *fv = 0;
    int dim = 0;
    ofstream fTemp;
    
    if (fichSalida != 0){
        fTemporal.append(fichSalida);
        fTemporal.append("TempFile");
        fTemp.open(fTemporal.c_str(), ofstream::out | ofstream::trunc);
        fTemp << "";
        fTemp.close();
        
    }

    int err = getdir(directorio, files);
    getDB(directorio);
    int headerDim = 0;
    
    if (err == -1){ // Ejecutar con una sola huella
        
        LlamaMetodos *llm = new LlamaMetodos(directorio, ajustar, realzar);
        ImgUtil::setCurrentFileID(directorio);
        
        ofstream f2;
        if (fichSalidaOk){            
            // Para vaciar el fichero en caso de que ya existiera.
            f2.open(fichSalida, ofstream::out | ofstream::trunc); // podría pensarse en abrir y cerrar fichero cada vez
            f2.close();            
        }

        if (llm->doMethod(directorio, metodo) == -1){delete llm; return;}
        fv = llm->getFV();
        dim = llm->getDim();
        headerDim = dim;

        maximos = MemUtil::createVector<float>(dim + 1, -10000);
        minimos = MemUtil::createVector<float>(dim + 1, 10000);
        tipos = MemUtil::createVector<int>(dim + 1);
        valoresNom = MemUtil::create2DMatrix<int>(dim + 1, MaxValNom + 1); /*Para cada atributo, 
        * la primera posición indica cuántos valores nominales tiene, si son 3,
        * van de la posición 1 a la 3*/

        string fich = directorio;
        
        if (!fichSalidaOk){           
            printFeatureVector(fv, dim, fich, metodo); 
        }
        else if (fichSalidaOk){           
            ofstream fTemp;
            fTemp.open(fTemporal.c_str(), ofstream::out | ofstream::app);
            printFeatureVector(fv, dim, fich, metodo, fTemp);   
            fTemp.close();
        }
        else{
            cout << "Ejecucion: <huella wsq o directorio> <metodo>  <ajuste (1) o redimension (0)> [fichero de salida]" << endl;
        }
        
        delete llm;     
    }
    else{
         
        cout << "Files to be processed: " << endl;
        for (int i = 0;i < files.size();i++) {
            cout << files[i] << endl;
        }
        
        string direc = directorio;
        getDB(directorio);
        direc.append("/");
        
        ofstream f2;
        if (fichSalidaOk){            
            // Para vaciar el fichero en caso de que ya existiera.
            f2.open(fichSalida, ofstream::out | ofstream::trunc); // podría pensarse en abrir y cerrar fichero cada vez
            f2.close();            
        }
        //** Identifica el formato de huella en el constructor
        string fich;
        fich.assign(direc + files[0]);

        LlamaMetodos *llm = new LlamaMetodos(fich, ajustar, realzar);
        ImgUtil::setCurrentFileID(fich);
        
        //                                           **//
        bool initialized = false;
        for (int i = 0; i < files.size(); i++){
            //Crear el path completo de donde está la huella
            fich.assign(direc + files[i]);

            ImgUtil::setCurrentFileID(fich);
            cout << "\t" << fich << endl;            

            if (llm->doMethod(fich, metodo) == -1){delete llm; return;}
            fv = llm->getFV();
            dim = llm->getDim();        

            if (((i == 0) && ((dim > 1) || (fv[0] != -1))) || (((dim > 1) || (fv[0] != -1)) && !initialized)) {
                maximos = MemUtil::createVector<float>(dim + 1, -10000);
                minimos = MemUtil::createVector<float>(dim + 1, 10000);
                tipos = MemUtil::createVector<int>(dim + 1);
                valoresNom = MemUtil::create2DMatrix<int>(dim + 1, MaxValNom + 1);/*Para cada atributo, 
                * la primera posición indica cuántos valores nominales tiene, si son 3,
                * van de la posición 1 a la 3*/
                headerDim = dim;
                initialized = true;
            }

            //Para rechazar las huellas que tengan -1 excepto las de Leung
            size_t foundLeung;
            string name = metodo;
            foundLeung = name.find("Leung");

            if (!fichSalidaOk && (fv[dim-1] != -1 || foundLeung != string::npos))                
                printFeatureVector(fv, dim, fich, metodo);
            else if (fichSalidaOk && (fv[dim-1] != -1 || foundLeung != string::npos)){

                ofstream fTemp;
                fTemp.open(fTemporal.c_str(), ofstream::out | ofstream::app);
                printFeatureVector(fv, dim, fich, metodo, fTemp);   
                fTemp.close();
            }
            llm->clear();
            fv = 0;
        }

//        if (fichSalidaOk)
//            f2.close();

        delete llm;
    }
   
    if (fichSalidaOk){
        ofstream f2;
        f2.open(fichSalida, ofstream::out | ofstream::app);
        printDataSetHeader(headerDim, metodo, f2);
        copyDataSet(f2);
        f2.close();
        
        remove(fTemporal.c_str());
    }  
    files.clear();
}

int EjecutarMetodo::extractFingerprintType(string fileName){
    
    size_t found;
    transform(fileName.begin(), fileName.end(),fileName.begin(), ::tolower );
    found = fileName.find("arch");
    if (found != string::npos) return 1;
    
    found = fileName.find("tented");
    if (found != string::npos) return 2;
    
    found = fileName.find("left");
    if (found != string::npos) return 4;
    
    found = fileName.find("right");
    if (found != string::npos) return 3;
    
    found = fileName.find("whorl");
    if (found != string::npos) return 5;

    return -1;
}
