/* 
 * \file        ImgUtil.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Juan Cerrón González (juan.cerron@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of ImgUtil namespace
 *
 * 
 */

#include "ImgUtil.h"

 int debug = 0;

namespace ImgUtil {
    
    bool DEBUG = false;
    float fAdjust = 1;
    bool doAdjust = true;
    bool isSfinge = false;
    string currentFileID = "";
    string currentDB = "";
    
    void setDebug(bool dbg){
        ImgUtil::DEBUG = dbg;
    }
    
    void setfAdjust(float fAdjust) {
        ImgUtil::fAdjust = fAdjust;
    }
    
    void setDoAdjust(bool doAdjust) {
        ImgUtil::doAdjust = doAdjust;        
    }
    
    void setIsSfinge(bool isSfinge) {
        ImgUtil::isSfinge = isSfinge;             
    }

    int writeGrayScaleImageCore(float **image, int h, int w, string fileName, int coreX, int coreY) {
        float **finger = MemUtil::create2DMatrix<float>(h, w);

        for (int i = 0; i < w; i++) {
            for (int j = 0; j < h; j++) {
                if (!(((j > coreY - 2) && (j < coreY + 2)) && ((i > coreX - 2) && (i < coreX + 2))))
                    finger[j][i] = (int) image[j][i];
                else {
                    finger[j][i] = 255;
                }
            }
        }
        ImgUtil::writeGrayScaleImage(finger, h, w, fileName, 1);

        MemUtil::freeMatrix(&finger);
    }

    int writeSPImage(float **image, int nRow, int nCol, vector<float> deltas, vector<float> cores, int blockSize, string fileName) {

        int **newDeltas = 0;
        if (deltas.size() > 0)
            newDeltas = MemUtil::create2DMatrix<int>(deltas.size(), 2);
        int posCounter = 0;
        for (int i = 0; i < deltas.size() / 2; i++) {
            newDeltas[i][0] = (deltas[posCounter] + 0.5) * blockSize;
            posCounter++;
            newDeltas[i][1] = (deltas[posCounter] + 0.5) * blockSize;
            posCounter++;
        }

        int **newCores = 0;
        if (cores.size() > 0)
            newCores = MemUtil::create2DMatrix<int>(cores.size(), 2);
        posCounter = 0;
        for (int i = 0; i < cores.size() / 2; i++) {
            newCores[i][0] = (cores[posCounter] + 0.5) * blockSize;
            posCounter++;
            newCores[i][1] = (cores[posCounter] + 0.5) * blockSize;
            posCounter++;
        }
        writeSPImage(image, nRow, nCol, newDeltas, deltas.size() / 2, newCores, cores.size() / 2, fileName);
        MemUtil::freeMatrix(&newCores);
        MemUtil::freeMatrix(&newDeltas);
        return 0;
    }
    
    int writeSPImage(float **image, int nRow, int nCol, vector<int> deltas, vector<int> cores, int blockSize, string fileName) {

        int **newDeltas = 0;
        if (deltas.size() > 0)
            newDeltas = MemUtil::create2DMatrix<int>(deltas.size(), 2);
        int posCounter = 0;
        for (int i = 0; i < deltas.size() / 2; i++) {
            newDeltas[i][0] = (deltas[posCounter] + 0.5) * blockSize;
            posCounter++;
            newDeltas[i][1] = (deltas[posCounter] + 0.5) * blockSize;
            posCounter++;
        }

        int **newCores = 0;
        if (cores.size() > 0)
            newCores = MemUtil::create2DMatrix<int>(cores.size(), 2);
        posCounter = 0;
        for (int i = 0; i < cores.size() / 2; i++) {
            newCores[i][0] = (cores[posCounter] + 0.5) * blockSize;
            posCounter++;
            newCores[i][1] = (cores[posCounter] + 0.5) * blockSize;
            posCounter++;
        }
        writeSPImage(image, nRow, nCol, newDeltas, deltas.size() / 2, newCores, cores.size() / 2, fileName);
        MemUtil::freeMatrix(&newCores);
        MemUtil::freeMatrix(&newDeltas);
        return 0;
    }
    int writeSPImage(float **image, int nRow, int nCol, int *deltas, int nDeltas, int *cores, int nCores, string fileName) {

        int **newDeltas = 0;
        if (nDeltas > 0)
            newDeltas = MemUtil::create2DMatrix<int>(nDeltas, 2);
        int posCounter = 0;
        for (int i = 0; i < nDeltas; i++) {
            newDeltas[i][0] = deltas[posCounter];
            posCounter++;
            newDeltas[i][1] = deltas[posCounter];
            posCounter++;
        }

        int **newCores = 0;
        if (nCores > 0)
            newCores = MemUtil::create2DMatrix<int>(nCores, 2);
        posCounter = 0;
        for (int i = 0; i < nCores; i++) {
            newCores[i][0] = cores[posCounter];
            posCounter++;
            newCores[i][1] = cores[posCounter];
            posCounter++;
        }

        writeSPImage(image, nRow, nCol, newDeltas, nDeltas, newCores, nCores, fileName);

        MemUtil::freeMatrix(&newCores);
        MemUtil::freeMatrix(&newDeltas);
        return 0;
    }

    int writeSPImage(float **image, int nRow, int nCol, int **deltas, int nDeltas, int **cores, int nCores, string fileName) {

        float **newImage = MemUtil::create2DMatrix<float>(nRow, nCol);

        float max = 0;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                newImage[i][j] = image[i][j];
                if (newImage[i][j] > max) {
                    max = newImage[i][j];
                }
            }
        }

        //Deltas = square
        for (int i = 0; i < nDeltas; i++) {
            int posI = deltas[i][0];
            int posJ = deltas[i][1];

            for (int aux = -4; aux <= 4; aux++) {
                if ((posI + aux > 0) && (posI + aux < nRow) && (posI - aux > 0) && (posI - aux < nRow) &&
                        (posJ + aux > 0) && (posJ + aux < nCol) && (posJ - aux > 0) && (posJ - aux < nCol)) {
                    newImage[posI - 4][posJ + aux] = 255;
                    newImage[posI + 4][posJ + aux] = 255;
                    newImage[posI + aux][posJ - 4] = 255;
                    newImage[posI + aux][posJ + 4] = 255;
                }
            }
        }

        //Cores = crosses
        for (int i = 0; i < nCores; i++) {
            int posI = cores[i][0];
            int posJ = cores[i][1];

            for (int aux = -4; aux <= 4; aux++) {
                if ((posI + aux > 0) && (posI + aux < nRow)) {
                    if ((posJ + aux > 0) && (posJ + aux < nCol)) 
                        newImage[posI + aux][posJ + aux] = 255;
                    
                    if ((posJ - aux > 0) && (posJ - aux < nCol)) 
                        newImage[posI + aux][posJ - aux] = 255;
                }
            }
        }

        writeGrayScaleImage(newImage, nRow, nCol, fileName, 1);

        MemUtil::freeMatrix(&newImage);
        return 0;
    }

    int writeGrayScaleImage(float **image, int nRow, int nCol, string fileName, float scale, bool **masc, int blockW) {

        ofstream f;
        fileName.append(".pgm");
        if (DEBUG)
            cout << "FingerprintImg: Writing image... " << fileName << endl;
        f.open(fileName.c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for gray scale image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P2" << endl;
        f << "# This is the bitmap" << endl;
        f << nCol << " " << nRow << endl;
        f << "255" << endl;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                if ( (masc != 0) && (!masc[(int)(i / blockW)][(int)(j / blockW)]))
                    f << (int) (image[i][j] * scale * 0.7) << " ";
                else
                    f << (int) (image[i][j] * scale) << " ";
            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        return 0;
    }

    int writeGrayScaleImage(int **image, int nRow, int nCol, string fileName, float scale) {
        
        ofstream f;
        fileName.append(".pgm");
        if (DEBUG)
            cout << "FingerprintImg: Writing image... " << fileName << endl;
        f.open(fileName.c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for gray scale image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P2" << endl;
        f << "# This is the bitmap" << endl;
        f << nCol << " " << nRow << endl;
        f << "255" << endl;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                f << (int) (image[i][j] * scale) << " ";

            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        return 0;
    }
    

    int writeGrayScaleImage(unsigned char **image, int nRow, int nCol, string fileName) {

        ofstream f;
        fileName.append(".pgm");
        if (DEBUG)
            cout << "FingerprintImg: Writing image... " << fileName << endl;
        f.open(fileName.c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for gray scale image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P2" << endl;
        f << "# This is the bitmap of the gabor filter image o eso creo" << endl;
        f << nCol << " " << nRow << endl;
        f << "255" << endl;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                f << (int) (image[i][j]) << " ";

            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        return 0;
    }

    int writeGrayScaleImage(bool **image, int nRow, int nCol, string fileName) {

        ofstream f;
        fileName.append(".pgm");
        if (DEBUG)
            cout << "FingerprintImg: Writing image... " << fileName << endl;
        f.open(fileName.c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for gray scale image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P1" << endl;
        f << "# This is the bitmap of the gabor filter image o eso creo" << endl;
        f << nCol << " " << nRow << endl;
        // f << "255" << endl;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                f << (int) (!image[i][j]) << " ";

            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        return 0;
    }

    int writeGrayScaleImage(int **image, int nRow, int nCol, string fileName) {

        ofstream f;
        fileName.append(".pgm");
        if (DEBUG)
            cout << "FingerprintImg: Writing image... " << fileName << endl;
        f.open(fileName.c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for gray scale image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P1" << endl;
        f << "# This is the bitmap of the gabor filter image o eso creo" << endl;
        f << nCol << " " << nRow << endl;
        // f << "255" << endl;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                f << (int) (image[i][j]) << " ";

            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        return 0;
    }
    
    int writeColorImage(int ***image, int nRow, int nCol, string fileName) {
        
        ofstream f;
        fileName.append(".ppm");
        if (DEBUG)
            cout << "FingerprintImg: Writing image... " << fileName << endl;
        f.open(fileName.c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for color image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P3" << endl;
        f << "# This is the bitmap" << endl;
        f << nCol << " " << nRow << endl;
        f << "255" << endl;
        for (int i = 0; i < nRow; i++) {
            for (int j = 0; j < nCol; j++) {
                for (int color = 0; color < 3; color++) {
                       f << image[i][j][color] << " ";
                }
            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        return 0;
    }

    void writeImageAndTrace(float **gImage, int **CI, string fileName, int w, int h, int Ws, int wCI, int hCI) {
        ofstream f;
        int i, j, x, y;
       
        int **Mat = MemUtil::create2DMatrix<int>(h, w);


        for (y = 0; y < hCI; y++) {
            for (x = 0; x < wCI; x++) {
                for (i = y * Ws; i < (y + 1) * Ws; i++) {
                    for (j = x * Ws; j < (x + 1) * Ws; j++) {
                        if ((i < h) && (j < w)) {
                            if (CI[y][x] > 0)
                                Mat[i][j] = CI[y][x];
                            else
                                Mat[i][j] = 0;
                        }
                    }

                }
            }
        }

        f.open(fileName.append(".pgm").c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for Trace image!" << endl;
        }

        /* Writing the file in PGM format */
        f << "P2" << endl;
        f << "# This is the bitmap of the Cores and deltas image" << endl;
        f << w << " " << h << endl;
        f << (int) 255 << endl;
        for (int i = 0; i < h; i++) {
            for (int j = 0; j < w; j++) {
                if (Mat[i][j] > 0)
                    f << (int) Mat[i][j] << " ";
                else
                    f << (int) gImage[i][j] << " ";

            }
            f << endl;
        }
        /* End of the PGM file format */
        f.close();

        MemUtil::freeMatrix(&Mat);
    }

    void xy_to_dmt(float **x, float **y, float **dg, float **mag, int w, int h, float *top_ret) {
        int i, j;
        float top, themag;
        static float a = 28.6479; /* 90./pi */

        for (i = 0, top = -1.; i < h; i++)
            for (j = 0; j < w; j++) {
                //   cout << i << ", " << j << endl;
                if (x[i][j] > 0.) {
                    if (y[i][j] >= 0.)
                        dg[i][j] = a * atan((double) (y[i][j] / x[i][j]));
                    else
                        dg[i][j] = a * atan((double) (y[i][j] / x[i][j])) + 180.;
                } else if (x[i][j] == 0.) {
                    if (y[i][j] > 0.)
                        dg[i][j] = 45.;
                    else if (y[i][j] < 0.)
                        dg[i][j] = 135.;
                    else
                        dg[i][j] = 0.;
                } else
                    dg[i][j] = a * atan((double) (y[i][j] / x[i][j])) + 90.;
                if ((themag = mag[i][j] = sqrt((double) (x[i][j] * x[i][j] +
                        y[i][j] * y[i][j]))) > top)
                    top = themag;
            }
        *top_ret = top;
    }

    void dmt_to_pic(float **dg, float **mag, int aw, int ah, float top, unsigned char *pic, int w, int h) {
        int i, j, iangle, ii, jj, ilen, jjlo, jjhi, iyow, jyow;
        float angle, ac, as, x0[WS][WS], y0[WS][WS], hurl;
        static int f = 1, bars[WS][WS][32][4];
        static float del = .09817 /* pi/32. */, fac = .177778 /* 32./180.*/;
        unsigned char *ptr, *tptr;
        float hwsp5;
        int jstp;

        if (f) {
            f = 0;
            if (HWS < 4 || HWS % 4)
                fprintf(stdout, "Blocksize (HWS) should be >= 4 and even divisible by 4\n");
            hwsp5 = (float) HWS - 0.5;
            jstp = HWS / 4;
            for (i = 0; i < WS; i++)
                for (j = 0; j < WS; j++) {
                    x0[i][j] = (float) j - hwsp5;
                    y0[i][j] = hwsp5 - (float) i;
                }
            for (iangle = 0, angle = 0.; iangle < 32; iangle++, angle += del) {
                ac = cos((double) angle);
                as = sin((double) angle);
                for (i = 0; i < WS; i++)
                    for (j = 0; j < WS; j++) {
                        hurl = ac * x0[i][j] + as * y0[i][j] + hwsp5;
                        jj = ((hurl >= 0.) ? hurl + .5 : hurl - .5);
                        hurl = as * x0[i][j] - ac * y0[i][j] + hwsp5;
                        ii = ((hurl >= 0.) ? hurl + .5 : hurl - .5);
                        jjlo = HWS - jstp;
                        jjhi = HWS + jstp - 1;
                        for (ilen = 0; ilen < 4; ilen++, jjlo -= jstp, jjhi += jstp)
                            bars[i][j][iangle][ilen] = (ii >= HWS - 1 && ii <= HWS
                                && jj >= jjlo && jj <= jjhi);
                    }
            }
        }
        for (i = 0; i < w * h; i++)
            pic[i] = 0;
        for (i = 0, iyow = WS; i < ah; i++, iyow += WS) {
            for (j = 0, jyow = WS; j < aw; j++, jyow += WS) {
                ilen = 3. * mag[i][j] / top + .5;
                if (ilen == 0)
                    continue;
                iangle = (int) (dg[i][j] * fac + .5);
                if (iangle == 32)
                    iangle = 0;
                tptr = pic + (iyow * w) + jyow;
                for (ii = 0; ii < WS; ii++) {
                    ptr = tptr;
                    for (jj = 0; jj < WS; jj++) {
                        if (bars[ii][jj][iangle][ilen])
                            *ptr = 255;
                        ptr++;
                    }
                    tptr += w;
                }
            }
        }
    }

    unsigned char *getOrientationImagePicture(float **avOrX, float **avOrY, int wIn, int hIn, int *wOut, int *hOut) {

        float **dg, **mag;
        unsigned char *bytes_pic;
        float top;
        int w, h;
        w = (WS + 1) * wIn + 2;
        h = (WS + 1) * hIn + 2;

        string s = "PCASYSManager: getOrientationImagePicture error allocating memory";
        malloc_uchar(&bytes_pic, w * h, (char *) s.c_str());
        malloc_dbl_flt(&dg, hIn, wIn, (char *) s.c_str());
        malloc_dbl_flt(&mag, hIn, wIn, (char *) s.c_str());


        xy_to_dmt(avOrX, avOrY, dg, mag, wIn, hIn, &top);
        dmt_to_pic(dg, mag, wIn, hIn, top, bytes_pic, w, h);

        free_dbl_flt(dg, hIn);
        free_dbl_flt(mag, hIn);

        *wOut = w;
        *hOut = h;

        return bytes_pic;

    }

    int writeOrImg(float **X, float **Y, int wOr, int hOr, string fileName) {

        unsigned char *orImage;
        ofstream f;
        int wOut, hOut;


        orImage = getOrientationImagePicture(X, Y, wOr, hOr, &wOut, &hOut);

        f.open(fileName.append(".pbm").c_str(), ofstream::out | ofstream::trunc);
        if (!f) {
            cout << "FingerprintImg: Error opening file for orientation image!" << endl;
            return -1;
        }

        /* Writing the file in PBM format */
        f << "P1" << endl;
        f << "# This is the bitmap of the orientation image" << endl;
        f << wOut << " " << hOut << endl;
        for (int i = 0; i < hOut; i++) {
            for (int j = 0; j < wOut; j++) {
                f << 1 - (int) orImage[i * wOut + j] / 255 << " "; //j * sw + i
            }
            f << endl;
        }
        /* End of the PBM file format */

        f.close();
        free(orImage);

        return 0;
    }

    float *computeSobel(float **img, int h, int w, float **gradX, float **gradY, float **grad) {

        float *MaxMin = MemUtil::createVector<float>(3); // [0] for max amplitude; [1] for min amplitude; [2] for the mean;
        MaxMin[0] = -10000000;
        MaxMin[1] = 25500000;
        MaxMin[2] = 0;

        /* compute SOBEL */
        for (int i = 1; i < h - 1; i++) {
            for (int j = 1; j < w - 1; j++) {
                gradX[i][j] = (img[i + 1][j + 1] + 2.0 * img[i][j + 1] + img[i - 1][j + 1] -
                        img[i + 1][j - 1] - 2.0 * img[i][j - 1] - img[i - 1][j - 1]);

                gradY[i][j] = -(img[i + 1][j - 1] + 2.0 * img[i + 1][j] + img[i + 1][j + 1] -
                        img[i - 1][j - 1] - 2.0 * img[i - 1][j] - img[i - 1][j + 1]);

                grad[i][j] = sqrt(gradX[i][j] * gradX[i][j] + gradY[i][j] * gradY[i][j]);
                MaxMin[2] += grad[i][j];

                if (MaxMin[0] < grad[i][j])
                    MaxMin[0] = grad[i][j];
                if (MaxMin[1] > grad[i][j])
                    MaxMin[1] = grad[i][j];

            }
        }

        MaxMin[2] = MaxMin[2] / ((h - 2)*(w - 2));

        return MaxMin;
    }
    
    void computeSobelGradients(float **img, int h, int w, float ***gradX, float ***gradY, float ***grad) {

        *gradX = MemUtil::create2DMatrix<float>(h,w);
        *gradY = MemUtil::create2DMatrix<float>(h,w);
        *grad = MemUtil::create2DMatrix<float>(h,w);
        float *aux = computeSobel(img, h, w, *gradX, *gradY, *grad);
        MemUtil::freeMatrix(&aux);
    }
    
    float **computeOrientationsFromGradients(float **gradX, float **gradY, int h, int w, int blockW, 
            float ***xCos, float ***ySin, int *orH, int *orW, bool normalize, float **coherence) {
        
        *orH = h / blockW;
        *orW = w / blockW;

        float **O = MemUtil::create2DMatrix<float>(*orH, *orW);
        *xCos = MemUtil::create2DMatrix<float>(*orH, *orW);
        *ySin = MemUtil::create2DMatrix<float>(*orH, *orW);
        
        
        float **Vx = MemUtil::create2DMatrix<float>(*orH, *orW);
        float **Vy = MemUtil::create2DMatrix<float>(*orH, *orW);
        float Gsx, Gsy;
        /* Compute Orientations */
        for (int i = 0; i < *orH; i++) {
            for (int j = 0; j < *orW; j++) {

                for (int u = i * blockW; u < (i + 1) * blockW; u++) {
                    for (int v = j * blockW; v < (j + 1) * blockW; v++) {
                        Gsx = 2.0 * gradX[u][v] * gradY[u][v];
                        Gsy = (gradX[u][v] * gradX[u][v] - gradY[u][v] * gradY[u][v]);
                        Vx[i][j] += Gsx;
                        Vy[i][j] += Gsy;
                        if (coherence != 0)
                           coherence[i][j] += sqrt(Gsx * Gsx + Gsy * Gsy);
                    }
                }
                // Coherence NOT NEEDED HERE!
                if (coherence != 0) {
                    coherence[i][j] = sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]) / coherence[i][j];
                    if (isnan(coherence[i][j]))
                        coherence[i][j] = 0;
                }
//Vx[i][j] /= (float) (blockW * blockW);
//Vy[i][j] /= (float) (blockW * blockW);
                float ang = 0.5 * atan2(Vx[i][j], Vy[i][j]);
                if (ang <= 0)
                    O[i][j] = ang + 0.5 * PI;
                else
                    O[i][j] = ang - 0.5 * PI;
                
                (*xCos)[i][j] = cos(2.0 * O[i][j]);
                (*ySin)[i][j] = sin(2.0 * O[i][j]);
                if (!normalize) {
                    float mag = sqrt(Vx[i][j] * Vx[i][j] + Vy[i][j] * Vy[i][j]);
                    (*xCos)[i][j] *= mag;
                    (*ySin)[i][j] *= mag;
                }
            }
        }
        
        MemUtil::freeMatrix(&Vx);
        MemUtil::freeMatrix(&Vy);
        return O;
    }

    

    int **binarizeImg(float **img, int h, int w, int umbral) {

        int **imgBin = MemUtil::create2DMatrix<int>(h, w);

        for (int i = 0; i < h; i++)
            for (int j = 0; j < w; j++)
                if ((img[i][j] < umbral) || isnan(img[i][j]))
                    imgBin[i][j] = 0;
                else
                    imgBin[i][j] = 1;

        return imgBin;
    }

    bool **imDilate(bool **img, int orH, int orW) {
        bool **imgDilated = MemUtil::create2DMatrix<bool>(orH, orW);

        /* Dilatation */
        for (int i = 0; i < orH; i++) {
            for (int j = 0; j < orW; j++) {
                if (!img[i][j]) {
                    bool dilatated = false;
                    for (int u = max(0, i - 1); u < min(orH, i + 2) && !dilatated; u++)
                        for (int v = max(0, j - 1); v < min(orW, j + 2); v++)
                            if (img[u][v]) {
                                imgDilated[i][j] = true;
                                dilatated = true;
                                break;
                            }
                } else
                    imgDilated[i][j] = true;
            }
        }
        return imgDilated;
    }

    bool **imErode(bool **img, int orH, int orW) {
        bool **imgEroded = MemUtil::create2DMatrix<bool>(orH, orW);

        /* Erosion */
        for (int i = 0; i < orH; i++) {
            for (int j = 0; j < orW; j++) {
                if (img[i][j]) {
                    bool erored = false;
                    imgEroded[i][j] = true;
                    for (int u = max(0, i - 1); u < min(orH, i + 2) && !erored; u++)
                        for (int v = max(0, j - 1); v < min(orW, j + 2); v++)
                            if (!img[u][v]) {
                                imgEroded[i][j] = false;
                                erored = false;
                                break;
                            }
                } else
                    imgEroded[i][j] = false;
            }
        }
        return imgEroded;


    }

    bool **imFill(bool **img, int orH, int orW, int nRounds) {
        bool **imgFilled = MemUtil::create2DMatrix<bool>(orH, orW);

        bool **imgCopy = MemUtil::copy2DMatrix(img, orH, orW);
        
        for (int k = 0; k < nRounds; k++) {
            /* Filling */
            for (int i = 0; i < orH; i++) {
                for (int j = 0; j < orW; j++) {
                    if (!imgCopy[i][j]) {
                        int count = 0;
                        for (int u = max(0, i - 1); u < min(orH, i + 2); u++)
                            for (int v = max(0, j - 1); v < min(orW, j + 2); v++) {
                                // cout << "u = " << u << " v = " << v << endl;
                                if (imgCopy[u][v]) {
                                    count++;
                                }
                            }
                        if (count > 4)
                            imgFilled[i][j] = true;
                        else
                            imgFilled[i][j] = false;
                    } else {
                        int count = 0;
                        for (int u = max(0, i - 1); u < min(orH, i + 2); u++)
                            for (int v = max(0, j - 1); v < min(orW, j + 2); v++) {
                                // cout << "u = " << u << " v = " << v << endl;
                                if (!imgCopy[u][v]) {
                                    count++;
                                }
                            }
                        if (count > 4)
                            imgFilled[i][j] = false;
                        else
                            imgFilled[i][j] = true;

                    }
                }
            }
            bool **aux = imgCopy;
            imgCopy = imgFilled;
            imgFilled = aux;
        }
        MemUtil::freeMatrix(&imgFilled);
        return imgCopy;
    }

    void smoothOrientations(float **O, float ***xCos, float ***ySin, int orH, int orW, int Wsmooth, bool normalizeLength, bool ***segmentation) {
        /* Smooth orientations */

        //   cout << "segmentation = " << segmentation << endl;
        int sWhalf = (Wsmooth - 1) / 2;
        int sWhalf2 = sWhalf;
        if (Wsmooth % 2 == 0)
            sWhalf2++;

        //        if (DEBUG) {
        //                ImgUtil::writeGrayScaleImage(segmentation, orH, orW, "ImgUtilSegmentation");
        //        }
        int nElements = 0;

        float **xCosSmooth = MemUtil::create2DMatrix<float>(orH, orW);
        float **ySinSmooth = MemUtil::create2DMatrix<float>(orH, orW);

        bool **segmentationAux = 0;
        if (segmentation != 0)
            segmentationAux = MemUtil::copy2DMatrix<bool>(*segmentation, orH, orW);

        // nElements = smoothOrientatioW * smoothOrientatioW;
        // cout << "suavizando! " << endl;
        bool apply = true;
        for (int i = 0; i < orH; i++) {
            for (int j = 0; j < orW; j++) {
                //cout << "es 0? = " << (segmentation==0) << endl;
                if ((segmentation == 0) || ((segmentation != 0) && ((*segmentation)[i][j]))) {
                    //cout << (*segmentation)[i][j] << " ";
                    //  cout << "eee" << endl;
                    nElements = 0;
                    apply = true;
                    for (int u = max(i - sWhalf, 0); u < min(i + sWhalf2 + 1, orH); u++) {
                        for (int v = max(j - sWhalf, 0); v < min(j + sWhalf2 + 1, orW); v++) {
                            xCosSmooth[i][j] += (*xCos)[u][v];
                            ySinSmooth[i][j] += (*ySin)[u][v];
                            nElements++;
                            //        cout << segmentation[u][v] << " ";
                            if ((segmentation != 0) && (!((*segmentation)[u][v])))
                                apply = false;
                        }
                        //     cout <<  endl;
                    }
                    //  cout <<  endl;

                    // cout << "apply = " << apply << endl;
                    if (apply) {
                        xCosSmooth[i][j] /= (float) nElements;
                        ySinSmooth[i][j] /= (float) nElements;
                        if (O != 0)
                            O[i][j] = 0.5 * atan2(ySinSmooth[i][j], xCosSmooth[i][j]);


                        if (normalizeLength && (O != 0)) {
                            //    cout << "normalize!" << endl;
                            xCosSmooth[i][j] = cos(2.0 * O[i][j]);
                            ySinSmooth[i][j] = sin(2.0 * O[i][j]);
                        }
                    } else {
                        segmentationAux[i][j] = false;
                        if (O != 0)
                            O[i][j] = 0;
                        xCosSmooth[i][j] = 0;
                        ySinSmooth[i][j] = 0;
                    }
                }
            }

        }
        MemUtil::freeMatrix(&(*xCos));
        MemUtil::freeMatrix(&(*ySin));
        *xCos = xCosSmooth;
        *ySin = ySinSmooth;
        if (segmentation != 0) {
            MemUtil::freeMatrix(&(*segmentation));
            *segmentation = segmentationAux;
        }
    }

    
    void setCurrentFileID(string id) {
        int indexIni = id.find_last_of("/") + 1;
        if (indexIni == 0)
            indexIni = id.find_last_of("\\") + 1;
        int indexFin = id.find(".wsq");
        if (indexFin == -1)
            indexFin = id.find(".WSQ");
        if (indexFin == -1 && DEBUG)
            cout << "ImgUtil: Fichero sin extensión .wsq" << endl;

        ImgUtil::currentFileID = id.substr(indexIni, indexFin);
    }
    
    void setCurrentDB(string id) {

        ImgUtil::currentDB = id;
    }

    string makeTestFileName(string infoName) {

        string finalName = "./TestImages/";
        finalName.append(ImgUtil::currentFileID);
        finalName.append("_");
        finalName.append(infoName);
        return finalName;
    }

    string makeTest2FileName(string infoName) {

        string finalName = "./Test2Images/";
        finalName.append(ImgUtil::currentFileID);
        finalName.append("_");
        finalName.append(infoName);
        return finalName;
	}

    string makeTest3FileName(string infoName){
        string finalName = "./Test2Images/Segmentations/";
        finalName.append(ImgUtil::currentFileID);
        finalName.append("_");
        finalName.append(infoName);
        return finalName;
    }
    
    string makeTest4FileName(string infoName){
        string finalName = "./Realces/";
        finalName.append(ImgUtil::currentDB);
        finalName.append("/");
        finalName.append(ImgUtil::currentFileID);
        finalName.append("_");
        finalName.append(infoName);
        return finalName;
    }

	float **createGaussianKernel(float sigma, int *gWout) {
                
        int gW = ceil(sigma * 6);
        float **gMasc = MemUtil::create2DMatrix<float>(gW, gW);
        float sigma2 = sigma * sigma;
        float x, y;
        int halfgW = gW / 2;
        for (int i = 0; i < gW; i++) {
            for (int j = 0; j < gW; j++) {
                x = i - halfgW;
                y = j - halfgW;
                gMasc[i][j] = 1. / (2. * PI * sigma2) * exp(- (x * x + y * y) / (2. * sigma2) );
            }
        }
        
        *gWout = gW;
        return gMasc;
    }
    
    float *create1DGaussianKernel(float sigma, int *gWout) {
                
        int gW = ceil(sigma * 6);
        float *gMasc = MemUtil::createVector<float>(gW);
        float sigma2 = sigma * sigma;
        float x;
        int halfgW = gW / 2;
        for (int i = 0; i < gW; i++) {
                x = i - halfgW;
                gMasc[i] = 1. / (sqrt(2. * PI) * sigma) * exp(- (x * x) / (2. * sigma2) );
        }
        *gWout = gW;
        return gMasc;
    }
    float **createGaussianKernel(float sigmaH, float sigmaW, int h, int w) {
        
        float **gMasc = MemUtil::create2DMatrix<float>(h, w);
        float sigma2 = sigmaH * sigmaW;
        float x, y;
        int halfH = h / 2;
        int halfW = w / 2;
        for (int i = 0; i < h; i++) {
            for (int j = 0; j < w; j++) {
                x = i - halfH;
                y = j - halfW;
                gMasc[i][j] = 1. / (2. * PI * sigma2) * exp(- ( ((x * x) / (2. * sigmaH * sigmaH)) + ((y * y) / (2. * sigmaW * sigmaW)) ) );
            }
        }
        
        return gMasc;
        
    }
    float **downSampleImg(float **img, int h, int w, int *hOut, int *wOut, int scale) {
        
        *hOut = h / scale;
        *wOut = w / scale;
        
        float scale2 = scale * scale;
        float **dsImg = MemUtil::create2DMatrix<float>(*hOut, *wOut);
        
        for (int i = 0; i < *hOut * scale; i++) {
            for (int j = 0; j < *wOut * scale; j++) {              
                dsImg[i / scale][j / scale] += img[i][j] / scale2;
                
            }
        }
        return dsImg;
    }

    
    /** This function computes the histogram of the image*/
    float *computeHistogram(float **img, int nRow, int nCol){
        int aux = 0;
        float *hist;
        
        cout << "Computing histogram" << endl;
        
        hist = MemUtil::createVector<float>(256);
        
        for (int i = 0; i < nRow; i++){
                for (int j = 0; j < nCol; j++){
                        aux = img[i][j];
                        hist[aux] += 1;
                }
        }
        cout << "Histogram computed: " << endl;
        if (DEBUG)
                MemUtil::showMatrix(hist,256);
        
        return hist;
    }
    
    /** This function segments the image at a certain threhold**/
    float **thresholdImage(float **image, int nRow, int nCol, float threshold){
        float **imgThresholded = MemUtil::create2DMatrix<float>(nRow, nCol);

        for (int i = 0; i < nRow ; i++)
            for (int j = 0; j < nCol; j++){
                imgThresholded[i][j] = image[i][j] > threshold ? 1 : 0;
            }

        return imgThresholded;
    }
    
    void fillBorderColor(float **imagen, int h, int w, int iniRow, int endRow, int iniCol, int endCol, float color){
        
        //fill top rows
        for (int i = 0; i < iniRow; i++)
            for (int j = 0; j < w; j++)
                imagen[i][j] = color;
        
        //fill bottom rows
        for (int i = endRow; i < h; i++)
            for (int j = 0; j < w; j++)
                imagen[i][j] = color;
        
        //fill left column
        for (int i = iniRow; i < endRow; i++ )
            for (int j = 0; j < iniCol; j++)
                imagen[i][j] = color;
        
        //fill right column
        for (int i = iniRow; i < endRow; i++)
            for (int j = endCol; j < w; j++)
                imagen[i][j] = color;
        
        if (DEBUG){
            string fileName2 = "BorderImageFillColor";
            ImgUtil::writeGrayScaleImage(imagen, h, w, fileName2, 1);
        }
                
    }
    
    void fillBorderExtending(float **imgExtend, float **fingerImg, int filterRadioX, int filterRadioY, int h, int w){

        for (int i = 0; i < filterRadioX; i++)
            for (int j = 0; j < filterRadioY; j++){
                imgExtend[i][j] = fingerImg[0][0];
                imgExtend[i][j + w + filterRadioY] = fingerImg[0][w-1];
                imgExtend[i + h + filterRadioX][j] = fingerImg[h-1][0];
                imgExtend[i + h + filterRadioX][j + w + filterRadioY] = fingerImg[h-1][w-1];
            }
        for (int i = 0; i < filterRadioX; i++)
            for (int j = filterRadioY; j < w + filterRadioY; j++){
                imgExtend[i][j] = fingerImg[0][j-filterRadioY];
                imgExtend[h + i + filterRadioX][j] = fingerImg[h-1][j-filterRadioY];
            }
        for (int j = 0; j < filterRadioY; j++)
            for (int i = filterRadioX; i < h + filterRadioX; i++){
                imgExtend[i][j] = fingerImg[i-filterRadioX][0];
                imgExtend[i][w + j + filterRadioY] = fingerImg[i-filterRadioX][w-1];
            }
        
        if (DEBUG){
            string fileName2 = "BorderImageFillExtending";
            ImgUtil::writeGrayScaleImage(imgExtend, h, w, fileName2, 1);
        }
    }

    float otsuThresholding(int *histogram, float *values, int numValues) {


        float minSTD = 100000;
        int bestTHR = -1;

        for (int i = 0; i < numValues; i++) {
            float mean1 = 0;
            float accNum1 = 0;
            for (int j = 0; j <= i; j++) {
                mean1 = mean1 + histogram[j] * values[j];
                accNum1 = accNum1 + histogram[j];
            }
            mean1 = mean1 / accNum1;

            float accStd1 = 0;
            for (int j = 0; j <= i; j++) {
                accStd1 = accStd1 + histogram[j]*((values[j] - mean1)*(values[j] - mean1));
            }
            accStd1 = accStd1 / accNum1;


            float mean2 = 0;
            float accNum2 = 0;
            for (int j = i + 1; j < numValues; j++) {
                mean2 = mean2 + histogram[j] * values[j];
                accNum2 = accNum2 + histogram[j];
            }
            mean2 = mean2 / accNum2;

            float accStd2 = 0;
            for (int j = i + 1; j < numValues; j++) {
                accStd2 = accStd2 + histogram[j]*((values[j] - mean2)*(values[j] - mean2));
            }
            accStd2 = accStd2 / accNum2;

            if (accStd1 + accStd2 < minSTD) {
                minSTD = accStd1 + accStd2;
                bestTHR = i;
            }


        }

        return values[bestTHR];

    }
    
    float directionDifference(float a1, float a2) {
        
        float dif = (a1 - a2);
        if (dif > PI)
            dif = dif - 2 * PI;
        else if (dif < -PI)
            dif = dif + 2 * PI;
        
        
        return abs(dif);
    }
    
    float directionMean(float a1, float a2) {
        
        
        return atan2(0.5 * (sin(a1) + sin(a2)), 0.5 * (cos(a1) + cos(a2)));
    }
    
}