/* 
 * \file        LlamaMetodos.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Juan Cerrón González (juan.cerron@unavarra.es) 
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The definition of LlamaMetodos methods
 *
 * 
 */

#ifndef LLAMAMETODOS_H
#define	LLAMAMETODOS_H


#include <string>
#include "PCASYSManager.h"

class LlamaMetodos {
    
public:
    LlamaMetodos();
    LlamaMetodos(string file, bool ajustar, bool realzar);
    ~LlamaMetodos();
    void clear();
    /**
     * Uses a string argument to execute a method.
     * \param file fingerprint
     * \param metodo method's name to execute
     * \return -1 if an error ocurred
     */
    int doMethod(string file, string metodo);
    
    /**
     * Call PCASYS method
     * \param file contains the fingerprint to work with
     */
    int doPCASYS(string file);
    
    /**
     * Call PCASYS method with fAdjust initialized for the correspondig fingerprint
     * \param file contains the fingerprint to work with
     */
    int doPCASYS2(string file);

    /**
     * Call JAIN method
     * \param file contains the fingerprint to work with
     */
    int doJAIN(string file);
    

    int doJAINPoincare(string file);
    /**
     * Call LEUNG method
     * \param file contains the fingerprint to work with
     * \param case2 realizes a enhancement returning a binarized image
     */
    int doLEUNG(string file, bool case2 = false);
    
    /**
     * Call LEUNG method with the segmentation from poincare
     * \param file contains the fingerprint to work with
     * \param case2 realizes a enhancement returning a binarized image
     */
    int doLEUNGPoincareSegmentation(string file, bool case2 = false);

    int doImageEnhancementHong(string file);

    /**
     * Call LIU method
     * \param file contains the fingerprint to work with
     */
    int doLIU(string file);

    /**
     * Call LIU method with the segmentation from poincare
     * \param file contains the fingerprint to work with
     */
    int doLIUSegmentationPoincare(string file);


    /**
     * Call LE method
     * \param file contains the fingerprint to work with
     */
    int doLE(string file);

    /**
     * Call LE method with the segmentation from poincare
     * \param file contains the fingerprint to work with
     */
    int doLESegmentationPoincare(string file);

    /**
     * Call NYONGESA method
     * \param file contains the fingerprint to work with
     */
    int doNYONGESA(string file);

    /**
     * Call NYONGESA method
     * \param file contains the fingerprint to work with
     */
    int doKARU(string file);

    /**
     * Call Zhang method
     * \param file contains the fingerprint to work with
     */
    int doZhang(string file);
    
    /**
     * Call Zhang method using the previous segmented and enhanced fingerprint
     * \param file contains the segmented and enhanced fingerprint to work with
     */
    int doZhangSegment(string file);


    /**
     * Call Poincare method
     * \param file contains the fingerprint to work with
     */
    int doPoincare(string file);

    /**
     * Call Cappelli method
     * \param file contains the fingerprint to work with
     */
    int doCappelli(string file);
    
    /**
     * Call Cappelli2 (ensemble model) method
     * \param file contains the fingerprint to work with
     */
    int doCappelli2(string file);
    
    
    /**
     * Call Hong method
     * \param file contains the fingerprint to work with
     */
    int doHongFE(string file);

    /**
     * Call NilssonBigun method
     * \param file contains the fingerprint to work with
     */
    int doNilssonBigun(string file);
    
    /**
     * Call ShahSastry method
     * \param file contains the fingerprint to work with
     */
    int doShahSastry(string file);
    
    
    int doLi(string file);
    
    
    int doLiSegmentationPoincare(string file);
    
    int doPark21(string file);
    
    int doPark105(string file);
    
    int doXuejunTanFE(string file);
    
    int doWangFE(string file);
    
    int doRejectJain(string file);
    int doRejectJainPoincare(string file);
    int doRejectHong(string file);
    
    int doLIUSegmentationPoincareWithJainReject(string file);
    int doZhangWithJainReject(string file);
    int doCappelli2WithJainReject(string file);
    int doRejectArea(string file);
    
    int doLIUSegmentationPoincareWithAreaReject(string file);
    int doZhangWithAreaReject(string file);
    int doJainPoincareWithAreaReject(string file);
    
    int doKaruWithJainReject(string file) ;
    int doKaruWithAreaReject(string file);
    
    int doGetEnhanced(string file);
    
    float *getFV() const;
    int getDim() const;
    
public:
    
    float *fv;
    int dim;
    PCASYSManager *pcasys;
    int err;

};

inline float *LlamaMetodos::getFV() const { return fv; }
inline int LlamaMetodos::getDim() const { return dim; }
#endif	/* LLAMAMETODOS_H */

