/* 
 * \file        MaskRegionsRelation.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for MaskRegionsRelation (Capelli99) class
 *
 * 
 */


#ifndef MASKREGIONSRELATION_H
#define	MASKREGIONSRELATION_H

#include "MaskRegion.h"

/**
 * \class MaskRegionsRelation
 *
 * The MaskRegionsRelation class implements the structure of a relation
 * between the regions of a dynamic mask mask used in Cappelli1999's algorithm
 */
class MaskRegionsRelation {
public:
    /* Default constructor */
    MaskRegionsRelation();
    
    /**
     * Constructor of a relation, which needs the two regions and their angle relation
     * \param P1 The first region
     * \param P2 The second region
     * \param ang The angle relation between them
     */
    MaskRegionsRelation(MaskRegion *P1, MaskRegion *P2, float ang);
    
    /* Default destructor */
    ~MaskRegionsRelation();
    
    /**
     * It returns the first region of the relation
     * \return The first region of the relation
     */
    MaskRegion *getFirstRegion() const;
    
    /**
     * It returns the second region of the relation
     * \return The second region of the relation
     */
    MaskRegion *getSecondRegion() const;
    
    /**
     * It returns the angle relation between both regions
     * \return The angle relation between both regions
     */
    float getRelationAngle() const;
    
private:

    float ang;          //!< The angle relation between both regions
    MaskRegion *P1;     //!< The first region of the relation
    MaskRegion *P2;     //!< The second region of the relation
};


inline MaskRegion *MaskRegionsRelation::getFirstRegion() const { return P1; }
inline MaskRegion *MaskRegionsRelation::getSecondRegion() const { return P2; }
    
inline float MaskRegionsRelation::getRelationAngle() const { return ang; }

#endif	/* MASKREGIONSRELATION_H */

