 /* 
 * \file        MathUtil.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of MathUtil 
 *
 * 
 */


#include <cstdlib>
#include <iostream>
#include <fstream>
#include <string>

#include "./Eigen/Eigenvalues"
#include "./Eigen/src/Eigenvalues/EigenSolver.h"

#include "MemUtil.hpp"

using Eigen::MatrixXd;
using Eigen::VectorXcd;
using Eigen::EigenSolver;

namespace MathUtil{
    
    float **makeCovariancesMatrix(float **vectors, int numVectors,int numDimensions){
        
        
        //1- Calculate the average values per dimension

        float *avgVector=MemUtil::createVector<float>(numDimensions,0);

        for (int i=0; i<numDimensions; i++){
            float acc=0;
            for (int j=0; j<numVectors; j++){
                acc=acc+vectors[j][i];
            }
            avgVector[i]=(float)acc/(float)numVectors;
        }

        
        //2- Declare & fill the intermediate (data-avg) matrix
        float **vectors2=MemUtil::create2DMatrix<float>(numVectors,numDimensions,0);
        
        for (int i=0; i<numDimensions; i++){
            for (int j=0; j<numVectors; j++){
                vectors2[j][i]=vectors[j][i]-avgVector[i];
            }
        }



        //3- Declare & calculate the intermediate matrix
        
        float **cmat=MemUtil::create2DMatrix<float>(numDimensions,numDimensions,0);


        for (int i=0; i<numDimensions; i++){
            for (int j=0; j<numDimensions; j++){

                float acc=0;

                for (int k=0; k<numVectors; k++){
                    acc=acc+vectors2[k][i]*vectors2[k][j];
                }

                cmat[i][j]=(float) acc/(float)(numVectors-1);
            }
        }

        MemUtil::freeMatrix(&avgVector);
        MemUtil::freeMatrix(&vectors2);
        
        return cmat;
    }
    
    
    float *getEigenvalues(float **mat, int numRows){

        //1- Create an Eigen::MatrixXd
        MatrixXd m(numRows,numRows);

        for (int i=0; i<numRows;i++){
            for (int j=0; j<numRows;j++){
                m(i,j) = mat[i][j];
            }
        }


        //2- Solve the system
        EigenSolver<MatrixXd> es(m);

        VectorXcd eivals = es.eigenvalues();

        //3-Declare & fill the output vector
        complex<float> *eigValues=MemUtil::createVector< complex<float> >(numRows);
        
        for (int i=0; i<numRows; i++){
            eigValues[i]=eivals(i);
        }

        float *eigenvaluesRealPart= MemUtil::createVector<float>(numRows);

        for (int i=0;i<numRows;i++){
            eigenvaluesRealPart[i]=eigValues[i].real();
        }

        MemUtil::freeMatrix(&eigValues);
        
        return eigenvaluesRealPart;

    }
}

