 /* 
 * \file        MathUtil.hpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Carlos Lopez Molina (carlos.lopez@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation and header file of MathUtil 
 *
 * 
 */


#ifndef MATHUTIL_H
#define	MATHUTIL_H
#include <cstdlib>

namespace MathUtil{
    
    /* Obtains the covariances matrix from a vector list of numdimensions.
     * The returned matrix has numDimensions x numDimensions elements.
     * \param vectors Matrix with as many rows (first param) as vectors and as many
     *                  cols (second param) as the number of dimensions
     * \param numVectors Number of vectors in the first argument
     * \param numDimensions Number of dimensions of the problem
     **/
     
    float **makeCovariancesMatrix(float **vectors, int numVectors,int numDimensions);
    
    float *getEigenvalues(float **mat, int numRows);
    
    template<typename T> float computeMatrixDifferences(T **matrix1, T **matrix2, int filas, int columnas){
        float suma = 0.0;
        
        for (int i = 0; i < filas; i++){
            for (int j = 0; j < columnas; j++)
                suma += abs(matrix1[i][j] - matrix2[i][j]);            
        }
        return suma;
        
    }
    
    template<typename T> float computeMatrixDifferences(T *matrix1, T *matrix2, int columnas){
        float suma = 0.0;
        
        for (int i = 0; i < columnas; i++)
                suma += abs(matrix1[i] - matrix2[i]);            
        
        return suma;
    }
    
    template<typename T> float computeMatrixDifferences(T ***matrix1, T ***matrix2, int filas, int columnas, int dimensions){
        float suma = 0.0;
        
        for (int k = 0; k < dimensions; k++)
            for (int i = 0; i < filas; i++)
                for (int j = 0; j < columnas; j++)
                        suma += abs(matrix1[k][i][j] - matrix2[k][i][j]);            
        
        return suma;
    }
}

#endif	/* MATHUTIL_H */

