 /* 
 * \file        MultiClassifierCappelli.cpp
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of MultiClassifierCappelli (Capelli02) class
 *
 * 
 */

#include "MemUtil.hpp"
#include "FingerprintImg.h"

#include "MultiClassifierCappelli.h"
#include "SPClassificationKaru.h"
#include "OrientationEnhancementCappelli.h"
#include "DirectionImagePartitioningCapelli.h"
#include "ParametersCappelli.h"


MultiClassifierCappelli::MultiClassifierCappelli() {
}

MultiClassifierCappelli::~MultiClassifierCappelli() {
}

float *MultiClassifierCappelli::run(string file, PCASYSManager *pcasys) {
    
    DirectionImagePartitioningCapelli *dipc = new DirectionImagePartitioningCapelli();
    float *fv = dipc->run(file, pcasys); // fv tiene 5 posiciones 
    
    FingerprintImg fi;
    
    if (ImgUtil::fAdjust != 1){  
        fi.setPCASYSManager(pcasys);  
        fi.readFingerprintImage(file);      
        fi.setSegmentedPCASYS(fi.getImage(), fi.getH(), fi.getW());
    }
    else{
        fi.setPCASYSManager(pcasys);
        fi.readFingerprintImage(file);
        fi.segment();
    }
   fi.enhance();
   fi.computeOrientations();
    
    SPClassificationKaru *spck = new SPClassificationKaru();
    spck->setOrientationData(fi.getYSin(),fi.getXCos(),fi.getOrH(),fi.getOrW(), 16);
    spck->setImageData(fi.getSegmentedImageFloat(),fi.getH(),fi.getW());
    
    vector<float> cores, deltas;
    spck->computeCoresAndDeltas(&cores, &deltas, true);
    int nCores = cores.size() / 2;
    if (ImgUtil::DEBUG) {
            if (nCores > 0) {
            string fileName = "referencePoint";
            cores[0] += 1.5;
            cores[1] += 1.5;        
            if (nCores > 1) {
                cores[2] += 1.5;
                cores[3] += 1.5;
            }
            ImgUtil::writeSPImage(fi.getSegmentedImageFloat(),fi.getH(),fi.getW(), deltas, cores, 16, fileName);
            cores[0] -= 1.5;
            cores[1] -= 1.5;        
            if (nCores > 1) {
                cores[2] -= 1.5;
                cores[3] -= 1.5;
            }
        }
    }
    
    if (nCores == 0)
       fi.computeCore_r92a();
    else {
        int coreI, coreJ;
        if (nCores == 2) {
            if (cores[0] < cores[2]) {
                coreI = (cores[0] + 1) * 16 + 8;
                coreJ = (cores[1] + 1) * 16 + 8;
            }
            else {
                coreI = (cores[2] + 1) * 16 + 8;
                coreJ = (cores[3] + 1) * 16 + 8;
            }
        }
        else {
            coreI = (cores[0] + 1) * 16 + 8;
            coreJ = (cores[1] + 1) * 16 + 8;
        }
            
       fi.setCore(coreI, coreJ);
    }

    
   fi.registerOrientations();    
    if (ImgUtil::DEBUG)
       fi.writeOrientationImage("registeredOr", 1);
    
    OrientationEnhancementCappelli *iec = new OrientationEnhancementCappelli(fi.getRegisteredYSin(),fi.getRegisteredXCos(),fi.getOrH(),fi.getOrW());
    iec->enhance(sigmaCappelli, lambdaCappelli);
    
    dimFV =fi.getOrH() * fi.getOrW() + 5;
    
    float *fvAll = MemUtil::createVector<float>(dimFV);
    for (int i = 0; i < 5; i++) {
        fvAll[i] = fv[i];
    }
    int count = 5;
    float **ySin = iec->getSin2Theta();
    float **xCos = iec->getcos2Theta();
    for (int i = 0; i < fi.getOrH(); i++) {
        for (int j = 0; j < fi.getOrW(); j++) {
            fvAll[count] = 0.5 * atan2(ySin[i][j], xCos[i][j]);
            count++;
        }
    }
    cout << endl;   
        
    MemUtil::freeMatrix(&fv);
    return fvAll;
}