/* 
 * \file        PCASYSManager.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Imaplementation for PCASYSManager  class
 *
 * 
 */



#include "MemUtil.hpp"


#include "PCASYSManager.h"


//#define ImgUtil::DEBUG 0
PCASYSManager::PCASYSManager() {
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: initializing..." << endl;                                                       
    remove("pcasys.out"); 
    
     /* Call initialization routine.  It reads parameters and data,
    and fopens demo images list for reading and outfile for writing. */
    string s = "PCASYSconfig/pcasys.prs";
    pcasys_init((char *)s.c_str(), &sgmnt_prs, &enhnc_prs, &rors_slit_range_thresh,
        &r92a_discard_thresh, &rgar_prs, &trnsfrm_nrows_use, &pnn_mlp, &pnn_prs,
        &mlp_prs, &pseudo_prs, &combine_clash_confidence, &protos_fvs,
        &protos_classes, &tranmat, &fp_demo_images_list, &fp_out);
    
    sw = WIDTH; sh = HEIGHT; wOr = 0; hOr = 0;
    sfgw = 0; sfgh = 0;
    wTf = 0; hTf = 0;
    wOr2 = 0;
    hOr2 = 0;
    
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: Initialized" << endl;   
}

PCASYSManager::PCASYSManager(int w, int h) {
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: initializing..." << endl;                                                       
    remove("pcasys.out"); 
    
     /* Call initialization routine.  It reads parameters and data,
    and fopens demo images list for reading and outfile for writing. */
    string s = "PCASYSconfig/pcasys.prs";
    pcasys_init((char *)s.c_str(), &sgmnt_prs, &enhnc_prs, &rors_slit_range_thresh,
        &r92a_discard_thresh, &rgar_prs, &trnsfrm_nrows_use, &pnn_mlp, &pnn_prs,
        &mlp_prs, &pseudo_prs, &combine_clash_confidence, &protos_fvs,
        &protos_classes, &tranmat, &fp_demo_images_list, &fp_out);
    
    sw = w; sh = h; wOr = 0; hOr = 0;
    sfgw = 0; sfgh = 0;
    wTf = 0; hTf = 0;
    wOr2 = 0;
    hOr2 = 0;
    
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: Initialized" << endl;   
}


PCASYSManager::~PCASYSManager() {
    
    if (ImgUtil::DEBUG)
        cout <<  "PCASYSManager: Destructing..." << endl;
//    // free(protos_fvs);        // No se libera ya que da error, aparentemente es porque el pcasys hace algo raro internamente
//    if (tranmat != 0)
//        delete (tranmat);
//    //cout << protos_classes[0] << endl;
//    //free(protos_classes);     // No se libera ya que da error, aparentemente es porque el pcasys hace algo raro internamente
    fclose(fp_demo_images_list);
    fclose(fp_out);
    
    if (ImgUtil::DEBUG)
        cout <<  "PCASYSManager: Destructed " << endl;
}

unsigned char **PCASYSManager::segment(unsigned char *image, int w, int h, unsigned char ***segmentedFg) {
    
    unsigned char **segras;
    
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: segmented image dimensions " << sw << "x" << sh << endl;
    
    sgmnt(image, w, h, &sgmnt_prs, &segras, sw, sh, segmentedFg, &sfgw, &sfgh);
    
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: correctly segmented " << endl;
    
    return segras;
}

unsigned char **PCASYSManager::enhance(unsigned char **image, int w, int h) {
    
    unsigned char **ehras;
        
    enhnc(image, &enhnc_prs, &ehras, w, h);    
    
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: correctly enhanced " << endl;
    
    return ehras;
}


char **PCASYSManager::computeOrientations(unsigned char **image, int w, int h, float ***avrors_x, float ***avrors_y) {
    
    char **pixelrors;
    
    /* Extract pixelwise ridge-orientations, and average them to
    produce a 28x30-vector grid of local orientations */
    // aw y ah vienen dados por w y h, y el tamaño del bloque para hacer la media!
    rors(image, w, h, rors_slit_range_thresh, &pixelrors, avrors_x, avrors_y, &wOr, &hOr);
        
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: Orientation image extracted " << endl;
        
    return pixelrors;
        
}

void PCASYSManager::computeCore_r92a(float **avOrX, float **avOrY, int *coreX, int *coreY) {
    
    /* Convert orientation grid to angles, and send them to r92
      registration program, which finds the "core" */
      r92a(avOrX, avOrY, wOr, hOr, r92a_discard_thresh, coreX, coreY);
    
}

void PCASYSManager::computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr, int *coreX, int *coreY) {
    
    /* Convert orientation grid to angles, and send them to r92
      registration program, which finds the "core" */
      r92a(avOrX, avOrY, wOr, hOr, r92a_discard_thresh, coreX, coreY);
    
}

void PCASYSManager::computeCore_r92a(float **avOrX, float **avOrY, int wOr, int hOr, int *coreX, int *coreY, int blockSize) {
    
    float ratioW = wOr / 32.;
    float ratioH = hOr / 32.;
    int orW;
    int orH;
    float **orX;
    float **orY;
    int **numBlocks = MemUtil::create2DMatrix<int>(32, 32);
    
    if ((ratioW > 1) && (ratioH > 1)){
        orH = 32;
        orW = 32;
        
        orX = MemUtil::create2DMatrix<float>(orH, orW);
        orY = MemUtil::create2DMatrix<float>(orH, orW);
        
        for (int i = 0; i < hOr; i++)
            for (int j = 0; j < wOr; j++){
                orX[(int)(i / ratioH)][(int)(j/ratioW)] += avOrX[i][j];
                orY[(int)(i / ratioH)][(int)(j/ratioW)] += avOrY[i][j];
                numBlocks[(int)(i / ratioH)][(int)(j/ratioW)]++;
            }
        
        for (int i = 0; i < orH; i++)
            for (int j = 0; j < orW; j++){
                if (numBlocks[i][j] > 0){
                    orX[i][j] /= numBlocks[i][j];
                    orY[i][j] /= numBlocks[i][j];
                }
            }
        
    }
    else if (ratioW > 1){
        orH = hOr;
        orW = 32;
        
        orX = MemUtil::create2DMatrix<float>(orH, orW);
        orY = MemUtil::create2DMatrix<float>(orH, orW);
        
        for (int i = 0; i < hOr; i++)
            for (int j = 0; j < wOr; j++){
                orX[i][(int)(j/ratioH)] += avOrX[i][j];
                orY[i][(int)(j/ratioH)] += avOrY[i][j];
                numBlocks[i][(int)(j/ratioH)]++;
            }
        
        for (int i = 0; i < orH; i++)
            for (int j = 0; j < orW; j++){
                if (numBlocks[i][j] > 0){
                    orX[i][j] /= numBlocks[i][j];
                    orY[i][j] /= numBlocks[i][j];
                }
            }
    }
    else if (ratioH > 1){
        orH = 32;
        orW = wOr;
        
        orX = MemUtil::create2DMatrix<float>(orH, orW);
        orY = MemUtil::create2DMatrix<float>(orH, orW);
        
        for (int i = 0; i < hOr; i++)
            for (int j = 0; j < wOr; j++){
                orX[(int)(i / ratioH)][j] += avOrX[i][j];
                orY[(int)(i / ratioH)][j] += avOrY[i][j];
                numBlocks[(int)(i / ratioH)][j]++;
            }
        
        for (int i = 0; i < orH; i++)
            for (int j = 0; j < orW; j++){
                if (numBlocks[i][j] > 0){
                    orX[i][j] /= numBlocks[i][j];
                    orY[i][j] /= numBlocks[i][j];
                }
            }
    }
    else{
        orH = hOr;
        orW = wOr;
        
        orX = MemUtil::copy2DMatrix<float>(avOrX, orH, orW);
        orY = MemUtil::copy2DMatrix<float>(avOrY, orH, orW);
    }

    /* Convert orientation grid to angles, and send them to r92
      registration program, which finds the "core" */
      r92a(orX, orY, orW, orH, r92a_discard_thresh, coreX, coreY);
      
      //FALTA REASIGNAR: avOrX y avOrY
      *coreX = (*coreX / 16.) * ratioW * blockSize;
      *coreY = (*coreY / 16.) * ratioH * blockSize;
      
      MemUtil::freeMatrix(&orX);
      MemUtil::freeMatrix(&orY);
      MemUtil::freeMatrix(&numBlocks);
      
      
}

void PCASYSManager::registerOrientations(char **pixelOr, int coreX, int coreY, float ***avOrRegX, float ***avOrRegY) {
    int dummyW, dummH; //notice that the dimension is always wOr and hOr
    
    rgar(pixelOr, sw, sh, coreX, coreY, &rgar_prs, avOrRegX, avOrRegY, &dummyW, &dummH);
    
}

float *PCASYSManager::extractOrientationFeatureVector(float **avOrRegX, float **avOrRegY) {
    float *featVect;
    
    if (ImgUtil::DEBUG)
        cout << "PCASYSManager: Extracting feature vector..." << endl;
    if(pnn_mlp == PNN_CLSFR) {
        wTf = pnn_prs.trnsfrm_cls;
        hTf = pnn_prs.trnsfrm_rws;
        if (ImgUtil::DEBUG)
                cout << "PCASYSManager: Configured as PNN, tfw = " << wTf << " tfh = " << hTf << endl;
    }
    else if(pnn_mlp == MLP_CLSFR) {
        wTf = mlp_prs.trnsfrm_cls;
        hTf = mlp_prs.trnsfrm_rws;
        if (ImgUtil::DEBUG)
                cout << "PCASYSManager: Configured as RNN, tfw = " << wTf << " tfh = " << hTf << endl;
    }

    featVect = (float *)malloc_ch(trnsfrm_nrows_use * sizeof(float));
    
    // tfh has the size of the feature vector, tfw the original size of the image
    trnsfrm(avOrRegX, avOrRegY, wOr, hOr, trnsfrm_nrows_use, tranmat, wTf, hTf, featVect);
    
    return featVect;
}


void PCASYSManager::computeOrientationsForPseudoRidge(char **pixelOr, float ***avOrX2, float ***avOrY2) {
    
    /* Take the pixelwise orientations made earlier and average them
    (without registration) to make a 58x62-orientation grid, for
    use by the pseudoridge analyzer. */
    ar2(pixelOr, sw, sh, avOrX2, avOrY2, &wOr2, &hOr2);
    
}

bool PCASYSManager::smoothOrientationsForPseudoRidge(unsigned char **segmentedFg, float **avOrX2, float **avOrY2) {
    
    bool found_conup;

    /* Follow pseudoridges through the print, looking for a
    concave-upward shape. */
    found_conup = pseudo(segmentedFg, sfgw, sfgh, avOrX2, avOrY2, wOr2, hOr2, &pseudo_prs);   

    return found_conup;
    
}

unsigned char **PCASYSManager::getPseudoRidges(unsigned char **segmentedFg, float **avOrX2, float **avOrY2) {
    
    unsigned char **pseudoRidges;
    
    int i, j, ii, jj;
    float **dg, **us, **vs;
    char **bad;
    double adouble;
    ofstream f;
    
    
  //  int wD, hD; // Dummy
    unsigned char *pict;
    string fileName = "pseudo";
    
    malloc_dbl_flt(&dg, hOr2, wOr2, (char *)"pseudo dg");
    malloc_dbl_flt(&us, hOr2, wOr2, (char *)"pseudo us");
    malloc_dbl_flt(&vs, hOr2, wOr2, (char *)"pseudo vs");
    malloc_dbl_char(&bad, hOr2, wOr2, (char *)"pseudo bad");

    pseudo_avrors2_xys2has(avOrX2, avOrY2, dg, wOr2, hOr2);
    for(ii = 0; ii < hOr2; ii++)
        for(jj = 0; jj < wOr2; jj++)
            if(avOrX2[ii][jj] * avOrX2[ii][jj] +
                    avOrY2[ii][jj] * avOrY2[ii][jj] < pseudo_prs.slthresh1
                    || !segmentedFg[ii + 1][jj + 1])
                bad[ii][jj] = 1;
            else {
                bad[ii][jj] = 0;
                adouble = dg[ii][jj] * D2R;
                us[ii][jj] = pseudo_prs.stepsize * cos(adouble);
                vs[ii][jj] = pseudo_prs.stepsize * (-sin(adouble));
            }
    pseudoRidges = printHasConup(us, vs, pseudo_prs.maxturn,
        pseudo_prs.max_tilt, pseudo_prs.min_side_turn,
        pseudo_prs.initi_s, pseudo_prs.initi_e, pseudo_prs.initj_s,
        pseudo_prs.initj_e, pseudo_prs.maxsteps_eachdir, bad, dg, wOr2, hOr2);
    
    
   // pict = getOrientationImagePicture(us, vs, wOr2, hOr2, &wD, &hD);
    
    f.open(fileName.append(".pbm").c_str(), ofstream::out | ofstream::trunc);
    if (!f) {
       cout << "FingerprintImg: Error opening file for orientation image!" << endl;
       return 0;
    }
    /* Writing the file in PBM format */
    f << "P1" << endl;
    f << "# This is the bitmap of the orientation image" << endl;
    f << sw << " " << sh << endl;
    for (i = 0; i < sh; i++) {
        for (j = 0; j < sw; j++) {
            f << 1 - (int)pseudoRidges[i][j]/255 << " "; //j * sw + i
        }
        f << endl;
    }
    /* End of the PGM file format */
    
    f.close();        
    //free(pict);
    
    
    free_dbl_flt(dg, hOr2);
    free_dbl_flt(us, hOr2);
    free_dbl_flt(vs, hOr2);
    free_dbl_char(bad, hOr2);

    
    
    return pseudoRidges;
    
}


unsigned char **PCASYSManager::printHasConup(float **us, float **vs, const int maxturn,
         const int max_tilt, const int min_side_turn, const int initi_s,
         const int initi_e, const int initj_s, const int initj_e,
         const int maxsteps_eachdir, char **bad, float **dg, const int w,
         const int h)
{
    
    unsigned char **pseudoRidges;
    int anint, initi, initj, i, ii = 0, j, k = 0, k2 = 0;
    int idir, flip, previ, prevj, prevflip;
    static int f = 1;
    float u, v, ri = 0.0, rj = 0.0, aturn, adg, newdg;
    static float *turn, *flowdg;
    
  static float *riarr, *rjarr;
  

    malloc_dbl_uchar(&pseudoRidges, sh, sw, (char *)"pseudo imagen");
  
    for (int h2 = 0; h2 < sh; h2++)
        memset(pseudoRidges[h2], (unsigned char) 0, sw);
    
    if(f) {
        f = 0;
        turn = (float *)malloc_ch(anint = 2 * maxsteps_eachdir * sizeof(float));
        flowdg = (float *)malloc_ch(anint);
        
    riarr = (float *)malloc_ch(anint += 2 * sizeof(float));
    rjarr = (float *)malloc_ch(anint);
    
    }
  // Esto son los bucles para pasar por todos los píxeles, los límites se establecen en los parámetros de pcasys
    for(initi = initi_s; initi <= initi_e; initi++) {
        for(initj = initj_s; initj <= initj_e; initj++) {
            if(bad[initi][initj])       // Si es malo no se empieza un pseudo-ridge
                continue;
            // Para cada una de las dos direcciones posibles del pixel...
            for(idir = 1; idir <= 2; idir++) {
                adg = dg[initi][initj]; // Se coge la orientación
                if(idir == 1) {         // 1ª dirección (noroeste por ejemplo)
                    u = .5 * us[initi][initj];  // 0.5 * coseno?
                    v = .5 * vs[initi][initj];
                    previ = initi;      // se guarda el pixel anterior/actual
                    prevj = initj;
                    prevflip = 0;       // flag para saber la dirección?
                }
                else {                  // 2ª dirección (suroeste por ejemplo)
                    
                    riarr[--ii] = ri;   // para dibujarlo!?
                    rjarr[ii] = rj;
                    
                    k2 = k;
                    u = -.5 * us[initi][initj];
                    v = -.5 * vs[initi][initj];
                    prevflip = 1;       // flag para saber que es la segunda tanda?
                }
                /* Empezamos desde el pixel initi, initj; con k = 0 -> k puede ser la posición del ridge!? (cuantos puntos has pasado siguiéndolo) */
                for(ri = initi, rj = initj, k = 0; k < maxsteps_eachdir; k++) {
                    ri += v;    // se suma o se resta el coseno (arriba o abajo
                    rj += u;
                    i = (ri >= 0. ? ri + .5 : ri - .5); // Se utiliza para redondear!!!! i, j tienen que ser int!
                    j = (rj >= 0. ? rj + .5 : rj - .5);
                    if(i < 0 || i >= h || j < 0 || j >= w || bad[i][j]) // si nos hemos salido o es pixel malo...
                        break;
                    /* A tener en cuenta si el giro es demasiado grande! */
                    if((aturn = (newdg = dg[i][j]) - adg) > 90.) {
                        flip = !prevflip;
                        aturn -= 180.;
                    }
                    else if(aturn < -90.) {
                        flip = !prevflip;
                        aturn += 180.;
                    }
                    else        // ha girado tanto como para invertir las direcciones? (pasar a - en vez de + o vice versa)
                        flip = prevflip;
                    
                    // Si ha girado de más no nos sirve!
                    if(fabs((double)aturn) > maxturn)
                        break;
                    if(idir == 1) {     // dirección 1, entonces guarda dsde el medio hasta la primera posición y ángulo negativo¿?
                        turn[ii = maxsteps_eachdir - 1 - k] = -aturn;
                        flowdg[ii] = (prevflip ? adg : adg + 180.);     // grados del flow, suma 180º si ha variado en la última ocasión
                        
                        riarr[++ii] = ri;
                        rjarr[ii] = rj;
                        
                        previ = i;
                        prevj = j;
                    }
                    else {              // dirección 2, entonces guarda dsde el medio hasta la última posición
                        turn[ii = maxsteps_eachdir + k] = aturn;
                        flowdg[ii] = (flip ? newdg + 180. : newdg);
                        
                        riarr[++ii] = ri;
                        rjarr[ii] = rj;
                        
                    }
                    adg = newdg;
                    if((prevflip = flip)) {
                        u = -us[i][j];
                        v = -vs[i][j];
                    }
                    else {
                        u = us[i][j];
                        v = vs[i][j];
                    }
                }
                if(k == maxsteps_eachdir) {
                    ri += v;
                    rj += u;
                }
            }
            
            riarr[++ii] = ri;
            rjarr[ii] = rj;
            
            anint = maxsteps_eachdir - k2;
            
            /*for (int h = 0; h < k2 + k + 2; h++)
                cout << " " << riarr[anint + h];
            cout << endl;*/
            for (int h2 = 0; h2 < k2 + k + 2; h2++) {
                int rii = (riarr[anint + h2] / hOr2)*sh; //(WS+1) + 2;
                int rjj = (rjarr[anint + h2] / wOr2)*sw; //*(WS+1) + 2;
                int iii = (rii >= 0. ? rii + .5 : rii - .5);
                int jjj = (rjj >= 0. ? rjj + .5 : rjj - .5);
           //     cout << "iii = " << iii << " jjj = "  << jjj << endl;
                if ((iii < sh) && (jjj < sw))
                        pseudoRidges[iii][jjj] =  (unsigned char) 255;
                           
            }
            //return pseudoRidges;

            if(path_has_conup(turn + anint, flowdg + anint, k2 + k, max_tilt, min_side_turn))
                return pseudoRidges;
        }
    }
    return pseudoRidges;
}


