/* 
 * \file        ParametersLe.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The parameters for the SymVarReferencePointLe algorithm (Le2012-Pattern recognition)
 *
 * 
 */

#ifndef PARAMETERSLE_H
#define	PARAMETERSLE_H

///////////////////////////////
//      LE/FINGER_CODE       //
///////////////////////////////
const int defNumSectorsLe = 16;   //!<  Default number of sectors for the original FingerCode
const int numBandsLe = 10;     //!<  Number of bands
const int bandWithLe = 10;   //!<  Number of pixels per band

const int numGaborFiltersLe = 8;             //!<  Number of filters (orientations) for the circular/original FingerCode
       
const int V0Le = 100;  //!<  Desired Variance     //!<  Normalization parameters
const int M0Le = 100;  //!<  Desired Mean

const double gaborFrecuencyLe = 0.1;             //!<  f, frecuency to construct the gabor filter; recommended value is 1 / K being K the distance between ridges, for 500dpi images, K = 10 pixels
const double gaborDXLe = 4.0;                    //!<  bandwidth for the Gaussian in X 
const double gaborDYLe = 4.0;                    //!<  bandwidth for the Gaussian in Y

const int gaborSupportXLe = 33;                //!<  Support of the Gabor Filter
const int gaborSupportYLe = 33;                //!<  Support of the Gabor Filter


const int centerDisplacementLe = 0;             //!< Displacement of the center for fingercode


///////////////////////////////////
//      LE/MEI ORIENTATION FIELD //
///////////////////////////////////

const int wLeMei[3] = {6, 12, 18};                      //!<  Block size of the orientation images
const float varThresholdLeMei[3] = {0.25, 0.18, 0.12};  //!<  variance threshold for the orientation block selection
const int minMagnitudeLeMei = 10;

///////////////////////////////////
//      LE PARAMETERS           //
///////////////////////////////////

const int Ws2Le = 7;                                    //!<  Window size of the complex filter
const float sigmaCLe = 1.5;                             //!<  sigma of the gaussian used to process the complex filter

const int NLe = 4;                                      //!<  LEngth of the vertical line used in VORIV computation
const int KLe = 5;                                      //!<  Size of the neighborhood used in VORIV feature computation


// const int nOrientations = 16;     //!<  k = numDefSectors en Le parameters!!
const int initialLengthLe = 4;                          //!<  Length of the window used to compute the orientation of the core point
const int widthLe = 5;                                  //!<  Width of the window used to compute the orientation of the core point
const int maxLengthLe = 15;                             //!<  Maximum length of the window used to compute the orientation of the core point

#endif	/* PARAMETERSLE_H */

