 /* 
 * \file        ParametersLeung.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The parameters for the algorithm of Leung  
 *
 * 
 */

#ifndef PARAMETERSLEUNG_H
#define	PARAMETERSLEUNG_H

///////////
// LEUNG //
///////////
// Constants for ModifiedFingerCode (if they vary from the original for easeness)
const int blockSizeLeung = 16;
const int pixelPerBLockLeung = 39;             //!<  Pixels per block (each block is a square of PIXELS_PER_BLOCK * PIXELS_PER_BLOCK pixels(
const int numBlocksLeung = 4;                    //!<  Number of blocks from the center block to the edges (numSectors = (NUM_BLOCKS * 2 + 1)^2)
                                        //!<  4 + 1 distance to the farthest block

const int numGaborFiltersModLeung = 12;        //!<  Number of filters (orientations) for the modified/square FingerCode
        
const int V0Leung = 6000; //!<  Desired Variance        // Normalization parameters for Leung
const int M0Leung = 0;    //!<  Desired Mean

const double gaborFrequencyLeung = 0.0769;      //!<  frequency to construct Leung gabor filters result of 1 / (leung_factor=13)
const double gaborDXLeung = 5.5;                      //!<  bandwidth for the Gaussian in X 
const double gaborDYLeung = 5.5;                      //!<  bandwidth for the Gaussian in Y
const double factorLeung = 1.5;                        //!<  Leung adjust factor for yprima^2

const int gaborSupportXLeung = 23;                //!<  Support of the Gabor Filter
const int gaborSupportYLeung = 17;                //!<  Support of the Gabor Filter

const int filterRadioLeung = 1;                  //!<  circle filter's radio to average an image

const int centerDisplacementLeung = 40;                 //!< displacement from the center downwards for fingercode

/////////////////////////////////////////////////
// HONG ENHANCEMENT PARAMETERS USED IN LEUNG
/////////////////////////////////////////////////

const int smoothOrientatioWLeung = 5;                        //!< Size of the smoothing window for the orientation map
const int ridgeFrequencyInterpolateWLeung = 7;               //!< Size of the window for the frequency interpolation
const int ridgeFrequencyInterpolateMeanLeung = 0;            //!< Mean of the gaussian for the frequency interpolation
const int ridgeFrequencyInterpolateVarianceLeung = 9;        //!< Variance of the gaussian for the frequency interpolation
const int ridgeFrequencySmoothWLeung = 7;                    //!< Size of the window for the smoothing of the frequency
const int blockVarianceWLeung = 5;                           //!< Size of the window for the smoothing of the block variance 

const int varianceThresholdLeung = 125;                      //!< Threshold for the segmentation using the variance blocks


const double ridgeFrequencyMinLeung = 1.0 / 3.0;             //!< Minimum ridge frequency allowed (it is the maximum!)
const double ridgeFrequencyMaxLeung = 1.0 / 25.0;            //!< Maximum ridge frequency allowed (it is the minimum!)


const int wLeungHong = 16;                              //!< Block size of the orientation map
const int lLeungHong = 32;                              //!< Block size of the window for the computation of the x-signature
const int gaborSupLeungHong = 21; // 11                 //!< Gabor filter support size for the enhancement of the fingerprint
const float deltaGaborLeungHong = 4.0;                  //!< Delta value for the gaussians of the gabor filter



const float alphaC1LeungRealce = 0.0197;                        //!< Cluster 1 alpha center
const float betaC1LeungRealce = 0.0008;                         //!< Cluster 1 beta center
const float varC1LeungRealce = 0.0736;                          //!< Cluster 1 variance center

const float alphaC2LeungRealce = 0.6114;                        //!< Cluster 2 alpha center
const float betaC2LeungRealce = 0.1447;                         //!< Cluster 2 beta center
const float varC2LeungRealce = 0.2923;                          //!< Cluster 2 variance center

const float alphaC3LeungRealce = 0.3051;                        //!< Cluster 3 alpha center
const float betaC3LeungRealce = 0.0106;                         //!< Cluster 3 beta center
const float varC3LeungRealce = 0.0634;                          //!< Cluster 3 variance center

const float alphaC4LeungRealce = 0.3142;                        //!< Cluster 4 alpha center
const float betaC4LeungRealce = 0.1922;                         //!< Cluster 4 beta center
const float varC4LeungRealce = 0.0724;                          //!< Cluster 4 variance center

const float alphaC5LeungRealce = 0.4690;                        //!< Cluster 5 alpha center
const float betaC5LeungRealce = 0.1661;                         //!< Cluster 5 beta center
const float varC5LeungRealce = 0.1666;                          //!< Cluster 5 variance center

const float alphaC6LeungRealce = 0.0993;                        //!< Cluster 6 alpha center
const float betaC6LeungRealce = 0.1852;                         //!< Cluster 6 beta center
const float varC6LeungRealce = 0.0099;                          //!< Cluster 6 variance center

const bool clusterLeungRealce[] = {false, true, true /*false for restrictive segmentation*/, true, true, false}; //!< False represents the clusters for background
                                                                  //!< True represents the clusters for fingerprint
  

#endif	/* PARAMETERSLEUNG_H */

