/* 
 * \file        ParametersZhang.h
 * \author      Juan Cerrón González (juan.cerron@unavarra.es)
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The definition of Zhang's parameters
 *
 * 
 */

#ifndef PARAMETERSZHANG_H
#define	PARAMETERSZHANG_H

////////////////////////////
// Orientation Block Size //
////////////////////////////
const int blockSizeZhang = 16;                  //!< Size to create the orientation image.

///////////////////////////////
// Gaussian Filter Parameters//
///////////////////////////////
const double sigmaZhang = 1.4;                  //!< Sigma for the gaussian filter
const int gFilterHZhang = 5;                    //!< Height of he gaussian filter
const int gFilterWZhang = 5;                    //!< Width of he gaussian filter

//////////////////////////
// Coherence Parameters //
//////////////////////////
const double thresholdFactorZhang = 0.18;       //!< Threshold factor adjust
const int maxLabelsZhang = 60;                  //!< Maximum connected components to find

//////////////////////////
// Smoothing Parameters //
//////////////////////////
const int blockLZhang = 5;                      //!< Smoothing window length  
const int blockWZhang = 3;                      //!< Smoothing window width
const float thresholdXZhang = 0;                //!< Threshold used to calculate Mu in smoothing

///////////////////////////
// Core & Delta Analysis //
///////////////////////////

const float blockArroundCoreLZhang = 4;             //!< Core and delta analysis block lenght
const float blockArroundCoreWZhang = 2;              //!< Core and delta analysis block width
const int curveWZhang = 2;                      //!< Size of the matrix with wich the core curve is searched
const int minNumBlockZhang = 0;                 //!< Minimum number of blocks needed to consider a core or delta correct
const operationModePoincare poincareModeZhang = increasingSmoothPoincare; //<! Operation mode for the poincare method

const int rulesToBeAppliedZhang = 5; //5; //!< Karu rules are applied
const bool normalizeZhang = true; //!< Whether to normalize the orientation map componenets in poincare or not

//////////////////////////////
// Eliminate Singular Point //
//////////////////////////////
const float minDistanceZhang = 10;//4;                 //!< Minimum distance to consider independent cores and deltas

const int thresholdTentedArchZhang = 600;            //!< s threshold to consider a right/left loop instead of a tented arch
        
const int pseudoRidgeSizeZhang = 150;                        //!< Number of pixels which are followed by the pseudo-ridge

const int smoothWPoincareZhang = 3;                      //!< Smoothing window length  poincare

//////////////////
// Segmentation //
//////////////////
const float segThresholdZhang = 0.75;                         //!< Percentage of background segmented blocks to consider a fingerprint rejected

//const bool combinarIEH_NP = true;                       //!< Boolean wich determines if Image enhancement Hong segmentation must be combined with Poincare segmentation

//////////////////
const int V0ZhangHong = 100;  // Desired Variance     // Normalization parameters
const int M0ZhangHong = 100;  // Desired Mean
#endif	/* PARAMETERSZHANG_H */

