/* 
 * \file        Thinning.cpp
 * \author      Jose Antonio Sanz Delgado (joseantonio.sanz@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * The implementation of Thinning (for Shah) class
 *
 * 
 */


#include "Thinning.h"
#include "ImgUtil.h"
#include "ParametersShah.h"
#include "MemUtil.hpp"



Thinning::Thinning() {
    h = -1;
    w = -1;
}

Thinning::Thinning(int h, int w) {
    this->h = h;
    this->w = w;
    FirstProp = 0; //MemUtil::create2DMatrix<float>(h, w);
    FirstMarked = 0; //MemUtil::create2DMatrix<float>(h, w);
    crossing = 0; //MemUtil::create2DMatrix<float>(h, w);
    spurs = 0; //MemUtil::create2DMatrix<float>(h, w);
    prop = 0; //MemUtil::create2DMatrix<float>(h, w);
    marked = 0; //MemUtil::create2DMatrix<float>(h, w);
    slimmed = 0; //MemUtil::create2DMatrix<float>(h, w);
}

Thinning::~Thinning() {
    MemUtil::freeMatrix(&prop); 
    MemUtil::freeMatrix(&FirstProp);
    MemUtil::freeMatrix(&crossing);
    MemUtil::freeMatrix(&spurs); 
    MemUtil::freeMatrix(&marked); 
    MemUtil::freeMatrix(&FirstMarked); 
    MemUtil::freeMatrix(&slimmed); 
}

void Thinning::crossingNumberComputation2(float **img){
    float one, two, three, four, five, six, seven, eight, suma;
    MemUtil::freeMatrix(&spurs);
    MemUtil::freeMatrix(&crossing);
    crossing = MemUtil::create2DMatrix<float>(h, w, 0.0);
    spurs = MemUtil::create2DMatrix<float>(h, w, 0.0);
    for (int i = 1; i < h - 2; i++){
        for (int j = 1; j < w - 2; j++){
            //if the pixel is white
            if (img[i][j] != 0.0){
                suma = 0.0;
                one = img[i][j-1];
                two = img[i-1][j-1];
                three = img[i-1][j];
                four = img[i-1][j+1];
                five = img[i][j+1];
                six = img[i+1][j+1];
                seven = img[i+1][j];
                eight = img[i+1][j-1];
                /*We compute the crossing number*/
                crossing[i][j] = (one != three) + (three != five) + (five != seven) + (seven != one) + 2*(((two > one)&&(two > three))+((four > three)&&(four > five))+((six > five)&&(six > seven))+((eight > seven)&&(eight > one)));
                //we sum the white pixels of its neighbours to detect the noise spurs
                suma = one + two + three + four + five + six + seven + eight;
                spurs[i][j] = suma;
            }
        }
    }

}

void Thinning::distancePropagation(float **img){
    float min;
    float **propAux = MemUtil::copy2DMatrix<float>(img, h, w);
    
    MemUtil::freeMatrix(&prop);
    prop = MemUtil::copy2DMatrix<float>(img, h, w);
    //top-down left-right pass
    for (int i = 1; i < h - 2; i++){
        for (int j = 1; j < w - 2; j++){
            if (img[i][j] == 1.0){
                min = 100.0;//img pixel's have their intensity belonging to {0, 1}
                for (int k = -1; k <= 1; k++){
                    if (propAux[i-1][j+k] < min){
                        min = propAux[i-1][j+k];
                    }
                }
                if (propAux[i][j-1] < min){
                    min = propAux[i][j-1];
                }
                propAux[i][j] = min+1.0;
            }
        }
    }
    //down-top right-left pass
    for (int i = h - 2; i > 1; i--){
        for (int j = w - 2; j > 1; j--){
            if (img[i][j] == 1.0){
                min = 100.0;
                for (int k = 1; k >= -1; k--){
                    if ((prop[i+1][j+k]+1.0) < min){
                        min = prop[i+1][j+k] + 1.0;
                    }
                }
                if ((prop[i][j+1]+1.0) < min){
                    min = prop[i][j+1]+1.0;
                }
                if (propAux[i][j] < min){
                    min = propAux[i][j];
                }
                prop[i][j] = min;
            }
        }
    }

    MemUtil::freeMatrix(&propAux);
}

void Thinning::mark(){
    bool max;
    
    MemUtil::freeMatrix(&marked);
    marked = MemUtil::create2DMatrix<float>(h, w, 0.0);
    for (int i = 1; i < h - 2; i++){
        for (int j = 1; j < w - 2; j++){
            max = true;
            for (int auxI = -1; auxI <= 1; auxI++){
                for (int auxJ = -1; auxJ <= 1; auxJ++){
                    if (prop[i+auxI][j+auxJ] > prop[i][j]){
                        max = false;
                        break; //if it is not the maximum we stop checking
                    }
                }
                if (!max){
                    break;//if it is not the maximum we stop checking
                }
            }
            if (max){
                marked[i][j] = prop[i][j];
            }
        }
    }

}

void Thinning::slimmerNSEW(float **img){
    bool keepSlimming = true;
    
    slimmed = MemUtil::copy2DMatrix<float>(img, h, w);
    
    /*We assign the value zero to the pixels in the rows and columns which are not treated*/
    for (int i = 0; i < h-1; i++){
        slimmed[i][0] = 0.0;
        slimmed[i][w-1] = 0.0;
    }
    for (int j = 0; j < w-1; j++){
        slimmed[0][j] = 0.0;
        slimmed[h-1][j] = 0.0;
    }
    
    
    float **imgAux = MemUtil::copy2DMatrix(slimmed, h, w);
    
    crossingNumberComputation2(slimmed);  
    distancePropagation(slimmed);
    FirstProp = MemUtil::copy2DMatrix(prop, h, w);
    mark();
    FirstMarked = MemUtil::copy2DMatrix(marked, h, w);
    while (keepSlimming){
        keepSlimming = false;
        
        /*NORTH POINTS*/
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(marked[i][j] == 0.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for north points to be deleted
                    if ((slimmed[i-1][j] == 0.0)&&(slimmed[i+1][j] == 1.0)){
                            imgAux[i][j] = 0.0;
                            keepSlimming = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);

        /*SOUTH POINTS*/
        crossingNumberComputation2(slimmed);
        distancePropagation(slimmed);
        mark();
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(marked[i][j] == 0.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for south points to be deleted
                    if ((slimmed[i-1][j] == 1.0)&&(slimmed[i+1][j] == 0.0)){
                            imgAux[i][j] = 0.0;
                            keepSlimming = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        
        /*EAST POINTS*/
        crossingNumberComputation2(slimmed);
        distancePropagation(slimmed);
        mark();
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(marked[i][j] == 0.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for east points to be deleted
                    if ((slimmed[i][j-1] == 1.0)&&(slimmed[i][j+1] == 0.0)){
                            imgAux[i][j] = 0.0;
                            keepSlimming = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        
        /*WEST POINTS*/
        crossingNumberComputation2(slimmed);
        distancePropagation(slimmed);
        mark();
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(marked[i][j] == 0.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for west points to be deleted
                    if ((slimmed[i][j-1] == 0.0)&&(slimmed[i][j+1] == 1.0)){    
                            imgAux[i][j] = 0.0;
                            keepSlimming = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        crossingNumberComputation2(slimmed);  
        distancePropagation(slimmed);
        mark();
    }
    MemUtil::freeMatrix(&slimmed);
    slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
    
    MemUtil::freeMatrix(imgAux);
    
    if (ImgUtil::DEBUG){
            ImgUtil::writeGrayScaleImage(slimmed, h, w, "SlimmedImage", 255);
        }
    
}


void Thinning::thinner2(){
    bool keepThinning =  true;
    float **imgAux = MemUtil::copy2DMatrix(slimmed, h, w);
    
    //We follow the process performed in the slimming phase but allowing the marked points to be deleted
    //Therefore we avoid to call the distancePropagation and mark functions for each direction
    while (keepThinning){
        keepThinning = false;
        
        /*NORTH POINTS*/
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        crossingNumberComputation2(slimmed);
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for north points to be deleted
                    if ((slimmed[i-1][j] == 0.0)&&(slimmed[i+1][j] == 1.0)){
                            imgAux[i][j] = 0.0;
                            keepThinning = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        
        /*SOUTH POINTS*/
        crossingNumberComputation2(slimmed);
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for south points to be deleted
                    if ((slimmed[i-1][j] == 1.0)&&(slimmed[i+1][j] == 0.0)){
                            imgAux[i][j] = 0.0;
                            keepThinning = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        
        /*EAST POINTS*/
        crossingNumberComputation2(slimmed);
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for east points to be deleted
                    if ((slimmed[i][j-1] == 1.0)&&(slimmed[i][j+1] == 0.0)){
                            imgAux[i][j] = 0.0;
                            keepThinning = true;
                        }
                }
            }
        }
        MemUtil::freeMatrix(&slimmed);
        slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
        
        /*WEST POINTS*/
        crossingNumberComputation2(slimmed);
        for (int i = 1; i < h - 2; i++) {
            for (int j = 1; j < w - 2; j++){
                //we check if the point has to be deleted only if it is one in the original image
                if ((slimmed[i][j] == 1.0)&&(crossing[i][j] == 2.0)&&(spurs[i][j] != 1.0)){
                    //we look for west points to be deleted
                    if ((slimmed[i][j-1] == 0.0)&&(slimmed[i][j+1] == 1.0)){    
                            imgAux[i][j] = 0.0;
                            keepThinning = true;
                        }
                }
            }
        }
    }
    MemUtil::freeMatrix(&slimmed);
    slimmed = MemUtil::copy2DMatrix(imgAux, h, w);
    
    MemUtil::freeMatrix(&imgAux);
    
    if (ImgUtil::DEBUG){
            ImgUtil::writeGrayScaleImage(slimmed, h, w, "ThinnedImage", 255);
        }

}

void Thinning::clean(){
    bool adjacent;
    int num;

    for (int i = 1; i < h - 2; i++){
        for (int j = 1; j < w - 2; j++){
            if ((slimmed[i][j] == 1.0)&&(spurs[i][j] == 0.0)){
                //it is an isolated pixel because it is white and it has not white neighbours
                slimmed[i][j] = 0.0;
            }
            if (spurs[i][j] == 1.0){
                //It is a line end
                    adjacent = false;
                    num = 0;
                    for (int auxI = -1; auxI <= 1; auxI++){
                        for (int auxJ = -1; auxJ <= 1; auxJ++){
                            if (spurs[i+auxI][j+auxJ] > 2.0){
                                adjacent = true;
                            }
                            if (adjacent)
                                break;
                        }
                        if (adjacent)
                            break;
                    }
                    if (adjacent){
                        //It has some neighbour having more than three white neighbours 
                        slimmed[i][j] = 0.0;
                    }
                }
        }
    }         
    if (ImgUtil::DEBUG){
            ImgUtil::writeGrayScaleImage(slimmed, h, w, "CleanedImage", 255);
        }    
}

void Thinning::purge2(){
    int dist;
    
    crossingNumberComputation2(slimmed);  
    distancePropagation(slimmed);
    mark();
    
    float **auxMark = MemUtil::create2DMatrix<float>(h, w);
    for (int i = 1; i < h-2 ; i++){
        for (int j = 1; j < w-2; j++){
            if (slimmed[i][j] == 1.0){
                //We mark the line ends whose first distance is one
                //These are the begining points to follow
                if ((spurs[i][j] == 1.0)&&(FirstProp[i][j] == 1.0)){
                    auxMark[i][j] = 1.0;
                }
            }
        }
    }
    
    MemUtil::freeMatrix(&marked);
    marked = MemUtil::create2DMatrix<float>(h, w);
    for (int i = 1; i < h-2 ; i++){
        for (int j = 1; j < w-2; j++){
            if ((slimmed[i][j] == 1.0)&&(auxMark[i][j] == 1.0)){
                    dist = 1;
                    marked[i][j] = 1.0;//We mark this pixel as visited and we follow the line
                    dist = followLine(i, j, dist);
                    marked[i][j] = 0.0;//We unmark this pixel to allow it to be followed in other lines
                }
            }
        }
    
    MemUtil::freeMatrix(&auxMark);
    
    if (ImgUtil::DEBUG){
            ImgUtil::writeGrayScaleImage(slimmed, h, w, "PurgeImage", 255);
        }
    
}

int Thinning::followLine(int posI, int posJ, int dist){
    int nlds = noiseLineDistanceShah * ImgUtil::fAdjust;
    
    for (int i = -1; i <= 1; i++){
        for (int j = -1; j <= 1; j++){
                // If it is a line (2 white neighbours) and it has not been visited before we keep following the line
                if ((spurs[posI+i][posJ+j] == 2.0)&&(marked[posI+i][posJ+j] == 0.0)){
                    dist += 1;
                    marked[posI+i][posJ+j] = 1.0; //we mark the pixel as visited and we follow the line
                    dist = followLine(posI+i, posJ+j, dist);
                    marked[posI+i][posJ+j] = 0.0;//We unmark this pixel to allow it to be followed in other lines
                }
                else if (((spurs[posI+i][posJ+j] == 1.0)||(spurs[posI+i][posJ+j] > 2.0))&&(marked[posI+i][posJ+j] == 0.0)){
                    //end line of the followed line (1: end line; >2 the pixel is joined to the skeleton)
                    dist += 1;
                    marked[posI+i][posJ+j] = 1.0; //we mark the pixel as visited
                    if (dist<nlds){
                        //the distance is lower to the allowed line distance so we deleted the pixel
                        slimmed[posI+i][posJ+j] = 0.0;
                    }
                    marked[posI+i][posJ+j] = 0.0;//We unmark this pixel to allow it to be followed in other lines
                }
        }
    }
    if (dist < nlds){
        //the distance is lower to the allowed line distance so we deleted the pixel
        slimmed[posI][posJ] = 0.0;
    }
    return dist; 
}

void Thinning::cleanNoise(){
    float cont;
    
    int sbns = sizeBlockNoiseShah * ImgUtil::fAdjust;
    int slide = sbns/2;
    
    float **aux = MemUtil::create2DMatrix<float>(h, w);
    
    for (int i = slide; i < (h-slide-1); i++){
        for (int j = slide; j < (w-slide-1); j++){
            //if the pixel is white
            if (slimmed[i][j] == 1.0){
                cont = 0.0;
                for (int posI = -1*slide; posI <= slide; posI++){
                    for (int posJ = -1*slide; posJ <= slide; posJ++){
                        //we count the number of white pixels among its neighbours
                        if (slimmed[i+posI][j+posJ] == 1.0){
                            cont++;
                        }
                    }
                }
                if (cont < sbns-1){
                    //if the number of white neighbours is lower than the size of the block
                    //we mark it to be deleted
                    aux[i][j] = 1.0;
                }
            }            
        }
    }

    for (int i = 0; i < h; i++){
        for (int j = 0; j < w; j++){
            //if the pixel is marked to be deleted we delete it
            if (aux[i][j] == 1.0){
                slimmed[i][j] = 0.0;
            }
        }
    }
    
    MemUtil::freeMatrix(&aux);
    
    if (ImgUtil::DEBUG){
            ImgUtil::writeGrayScaleImage(slimmed, h, w, "FinalThinningImage", 255);
        }
}

int Thinning::getH(){
     return h;
 }

int Thinning::getW(){
    return w;
}

float **Thinning::getImg(){
    return slimmed;
}

float **Thinning::getMark(){
    return marked;
}



