/* 
 * \file        Thinning.h
 * \author      Mikel Galar Idoate (mikel.galar@unavarra.es)
 * \author      Jose Antonio Sanz Delgado (joseantonio.sanz@unavarra.es)
 * \version     1.0
 * 
 * \section     DESCRIPTION
 * 
 * Header file for Thinning class
 *
 * 
 */

#ifndef THINNING_H
#define	THINNING_H

class Thinning {
public:
    /*Default constructor*/
    Thinning();
    
    /**
     * It constructs an instance of the class to be processed with the following
     * parameters
     * \param image fingerprint image to extract the feature vector
     * \param h height of the image
     * \param w width of the image
     */
    Thinning(int h, int w);
    
     /*Default destructor*/
    virtual ~Thinning();
    
    /**
     * This function computes the crossing number of a given binarized image
     * parameters
     * \param img The image to be processed
     */
    void crossingNumberComputation2(float **img);
    
    /**
     * This function performs the distance propagation step of the thinning algorithm
     * parameters
     * \param img The binary image to be processed
     */
    void distancePropagation(float **img);
    
    /**
     * This function marks the pixels having a distance equal or greatter than its neighbours
     */
    void mark();
    
    /**
     * This function performs the slimming process: it reduces the objects to its skeleton
     */
    void slimmerNSEW(float **img);
    
    /**
     * This function reduces the skeleton carried out by means of the slimmerNSEW to unit length
     */
    void thinner2();
    
    /**
     * This function deletes both noisy lines joined to the skeleton and isolated lines 
     * The deleted lines have a length lower than NoiseLineDistance (default 15)
     */
    void purge2();
    
    /**
     * This is a recursive function to follow lines used in the purge process
     * parameters
     * /param posI Position of the pixel which calls recursively the function (height)
     * /param posJ Position of the pixel which calls recursively the function (width)
     * /param dist distance of the line followed until the call
     * /return the distance of the followed line
     */
    int followLine(int posI, int posJ, int dist);
    
    /**
     * This function cleans noise spurs and isolated pixels
     */
    void clean();
    
    /**
     * This function deletes pixels whose number of white neighbours is lower than sizeBlockNoise-1 
     * (default sizeBlockNoise = 9)
     */
    void cleanNoise();
    
    /**
     * This functions returns the height of the image
     */
    int getH();
    
    /**
     * This functions returns the widht of the image
     */
    int getW();
    
    /**
     * This functions returns the thinned image (slimmed variable)
     */
    float **getImg();
    
    /**
     * This functions returns the marked image 
     */
    float **getMark();
    
    
    
private:

    int h;              //!< Height of the image
    int w;              //!< Width of the image
    float **prop;       //!< Image for the distace propagation phase
    float **crossing;       //!< Image containing the crossing number of each pixel
    float **spurs;       //!< Image for the propagation phase
    float **marked;       //!< Image for the second step ot the thinnign algorithm (mark)
    float **slimmed;       //!< Image for the third step ot the thinnign algorithm (slim)
    float **FirstProp;       //!< Image to store the first distace propagation (for the purge process)
    float **FirstMarked;       //!< Image to store the first marked image (for the purge process)
};

#endif	/* THINNING_H */

